#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
macroparameterseditor.py: 
"""
import sys, inspect, glob

from taurus.external.qt import Qt
from taurus.core.util.singleton import Singleton
from taurus.qt.qtgui.resource import getThemeIcon

from sardana.taurus.core.tango.sardana import macro
from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.delegate import ParamEditorDelegate


class MacroParametersEditor(object):

    def __init__(self):
        pass
#        self._macroModel = None

#    def setMacroModel(self, model):
#        self._macroModel = model
#        self.connect(self._macroModel, Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), self.onDataChanged)
#        self.connect(self._macroModel, Qt.SIGNAL("modelReset()"), self.onModelReset)
#
#    def onDataChanged(self):
#        pass
#
#    def onModelReset(self):
#        self.onDataChanged()

class StandardMacroParametersEditor(Qt.QWidget, MacroParametersEditor):

    def __init__(self, parent=None, macroNode=None):
        Qt.QWidget.__init__(self, parent)
        self.initComponents()

    def initComponents(self):
        self.setLayout(Qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.tree = MacroParametersTree(self)
        self.delegate = ParamEditorDelegate(self.tree)
        self.tree.setItemDelegate(self.delegate)
        self.tree.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Expanding)
        self.layout().addWidget(self.tree)

        actionLayout = Qt.QVBoxLayout()
        actionLayout.setContentsMargins(0, 0, 0, 0)
        addButton = Qt.QToolButton()
        addButton.setDefaultAction(self.tree.addAction)
        actionLayout.addWidget(addButton)
        deleteButton = Qt.QToolButton()
        deleteButton.setDefaultAction(self.tree.deleteAction)
        actionLayout.addWidget(deleteButton)
        moveUpButton = Qt.QToolButton()
        moveUpButton.setDefaultAction(self.tree.moveUpAction)
        actionLayout.addWidget(moveUpButton)
        moveDownButton = Qt.QToolButton()
        moveDownButton.setDefaultAction(self.tree.moveDownAction)
        actionLayout.addWidget(moveDownButton)
        spacerItem = Qt.QSpacerItem(0, 0, Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Expanding)
        actionLayout.addItem(spacerItem)

        self.layout().addLayout(actionLayout)

    def setModel(self, model):
        self.tree.setModel(model)
        self.tree.expandAll()

    def macroNode(self):
        return self.tree.macroNode()

    def setMacroNode(self, macroNode):
        self.tree.setMacroNode(macroNode)

class MacroParametersTree(Qt.QTreeView):

    def __init__(self, parent=None, designMode=False):
        Qt.QTreeView.__init__(self, parent)
        self.setSelectionBehavior(Qt.QTreeView.SelectItems)
        self.setRootIsDecorated(False)
#        self.setTabKeyNavigation(True)
        self.setEditTriggers(Qt.QAbstractItemView.AllEditTriggers)

        self.addAction = Qt.QAction(getThemeIcon("list-add"), "Add new repetition", self)
        self.connect(self.addAction, Qt.SIGNAL("triggered()"), self.onAddRepeat)
        self.addAction.setToolTip("Clicking this button will add new repetition to current parameter.")

        self.deleteAction = Qt.QAction(getThemeIcon("list-remove"), "Remove repetition", self)
        self.connect(self.deleteAction, Qt.SIGNAL("triggered()"), self.onDelRepeat)
        self.deleteAction.setToolTip("Clicking this button will remove current repetition.")

        self.moveUpAction = Qt.QAction(getThemeIcon("go-up"), "Move up", self)
        self.connect(self.moveUpAction, Qt.SIGNAL("triggered()"), self.onUpRepeat)
        self.moveUpAction.setToolTip("Clicking this button will move current repetition up.")

        self.moveDownAction = Qt.QAction(getThemeIcon("go-down"), "Move down", self)
        self.connect(self.moveDownAction, Qt.SIGNAL("triggered()"), self.onDownRepeat)
        self.moveDownAction.setToolTip("Clicking this button will move current repetition down.")

        self.disableActions()


    def disableActions(self):
        self.addAction.setEnabled(False)
        self.deleteAction.setEnabled(False)
        self.moveUpAction.setEnabled(False)
        self.moveDownAction.setEnabled(False)

    def manageActions(self, currentIndex):
        self.disableActions()
        if currentIndex is None:
            return
        node = self.model().nodeFromIndex(currentIndex)
        if isinstance(node, macro.RepeatNode):
            self.deleteAction.setEnabled(not node.parent().isReachedMin())
            self.addAction.setEnabled(False)
            self.moveUpAction.setEnabled(node.isAllowedMoveUp())
            self.moveDownAction.setEnabled(node.isAllowedMoveDown())
        elif isinstance(node, macro.RepeatParamNode):
            self.addAction.setEnabled(not node.isReachedMax())
            self.deleteAction.setEnabled(False)


    def currentChanged(self, current, previous):
        self.manageActions(current)
        Qt.QTreeView.currentChanged(self, current, previous)

    #def focusInEvent(self, event):
    #    reason = event.reason()
    #    if (reason == Qt.Qt.TabFocusReason) | (reason == Qt.Qt.BacktabFocusReason):
    #        if reason == Qt.Qt.TabFocusReason:
    #            idx = self.forwardIdx(0, 1, Qt.QModelIndex())
    #        elif reason == Qt.Qt.BacktabFocusReason:
    #            idx = self.backwardIdx(len(self.root()) - 1, 1, Qt.QModelIndex())
    #        self.setCurrentIndex(idx)
    #        self.edit(idx)
    #    else:
    #        Qt.QTreeView.focusInEvent(self, event)
    #
    #def forwardIdx(self, row, col, parentIdx):
    #    try:
    #        proposalIdx = self.model().index(row, col, parentIdx)
    #    except AssertionError:
    #        if parentIdx.row() == -1:
    #            return Qt.QModelIndex()
    #        grandParentIdx = parentIdx.parent()
    #        return self.forwardIdx(parentIdx.row() + 1, col, grandParentIdx)
    #
    #    proposalNode = self.model().nodeFromIndex(proposalIdx)
    #
    #    if isinstance(proposalNode, macro.SingleParamNode):
    #        return proposalIdx
    #    elif isinstance(proposalNode, macro.RepeatNode):
    #        return self.forwardIdx(0, 1, proposalIdx)
    #    elif isinstance(proposalNode, macro.RepeatParamNode):
    #        if len(proposalNode) > 0:
    #            return self.forwardIdx(0, 1, proposalIdx)
    #        else:
    #            return self.forwardIdx(row + 1, col, proposalIdx)
    #    elif not proposalIdx.isValid():
    #        proposalIdx = parentIdx.sibling(parentIdx.row()+1, 0)
    #        if proposalIdx.isValid():
    #            proposalIdx = proposalIdx.child(0,1)
    #        else:
    #            while not proposalIdx.isValid():
    #                parentIdx = parentIdx.parent()
    #                if not parentIdx.isValid():
    #                    return Qt.QModelIndex()
    #                proposalIdx = parentIdx.sibling(parentIdx.row()+1, 1)
    #
    #        return proposalIdx
    #
    #    elif isinstance(proposalNode, macro.MacroNode):
    #        ##self.model().setRoot(proposalNode)
    #        return self.forwardIdx(0,1,proposalIdx)
    #
    #def backwardIdx(self, row, col, parentIdx):
    #    try:
    #        proposalIdx = self.model().index(row, col, parentIdx)
    #    except AssertionError:
    #        if parentIdx.row() == -1:
    #            return Qt.QModelIndex()
    #        grandParentIdx = parentIdx.parent()
    #        return self.backwardIdx(parentIdx.row() - 1, col, grandParentIdx)
    #    proposalNode = self.model().nodeFromIndex(proposalIdx)
    #    if isinstance(proposalNode, macro.SingleParamNode):
    #        return proposalIdx
    #    elif isinstance(proposalNode, macro.RepeatNode):
    #        return self.backwardIdx(self.model().rowCount(proposalIdx) - 1, 1, proposalIdx)
    #    elif isinstance(proposalNode, macro.RepeatParamNode):
    #        return self.backwardIdx(self.model().rowCount(proposalIdx) - 1, 1, proposalIdx)
    #
    #    elif not proposalIdx.isValid():
    #        proposalIdx = parentIdx.sibling(parentIdx.row()-1, 0)
    #        if proposalIdx.isValid():
    #            tempRow = 0
    #            proposalIdx = proposalIdx.child(tempRow,1)
    #            while proposalIdx.sibling(tempRow+1, 1).isValid():
    #                proposalIdx = proposalIdx.sibling(tempRow+1, 1)
    #                tempRow +=1
    #        else:
    #            while not proposalIdx.isValid():
    #                parentIdx = parentIdx.parent()
    #                if not parentIdx.isValid():
    #                    return Qt.QModelIndex()
    #                proposalIdx = parentIdx.sibling(parentIdx.row()-1, 1)
    #
    #        return proposalIdx
    #
    #def moveCursor (self, cursorAction, modifiers):
    #    ix=self.currentIndex()
    #    self.manageActions(ix)
    #    (col, row, parentIdx)=(ix.column(), ix.row(), ix.parent())
    #    #to start from second column
    #    if col == -1 and row == -1:
    #        if cursorAction == Qt.QAbstractItemView.MoveNext:
    #            return self.forwardIdx(0, 1, parentIdx)
    #        elif cursorAction == Qt.QAbstractItemView.MovePrevious:
    #            return self.backwardIdx(self.model().rowCount(parentIdx) - 1, 1, parentIdx)
    #    if (cursorAction == Qt.QAbstractItemView.MoveNext and
    #        modifiers == Qt.Qt.NoModifier):
    #        #This condition in case we start tabbing with cursor on first column
    #        if col == 0:
    #            currentNode = self.model().nodeFromIndex(ix)
    #            if isinstance(currentNode, macro.SingleParamNode):
    #                nextIdx = self.forwardIdx(row, 1, parentIdx)
    #            else:
    #                nextIdx = self.forwardIdx(0, 1, ix)
    #        else:
    #            nextIdx = self.forwardIdx(row + 1, 1, parentIdx)
    #        #this condition in case there is no next index and we want to pass focus
    #        #to next widget in parent obj
    #
    #        if nextIdx == "term":
    #            self.focusNextPrevChild(True)
    #            return Qt.QModelIndex()
    #
    #        if not nextIdx.isValid():
    #            self.parent().focusNextChild()
    #        #this condition in case the next index is valid and we want to
    #        #refresh state of buttons
    #        else:
    #            self.manageActions(nextIdx)
    #        return nextIdx
    #
    #    elif (cursorAction == Qt.QAbstractItemView.MovePrevious and
    #        modifiers == Qt.Qt.NoModifier):
    #        backwardIdx = self.backwardIdx(row - 1, 1, parentIdx)
    #        #this contion in case there is no previous index and we want to pass focus
    #        #to previous widget in parent obj
    #        if backwardIdx == "term":
    #            self.focusNextPrevChild(False)
    #            return Qt.QModelIndex()
    #
    #        if not backwardIdx.isValid():
    #            self.parent().focusPreviousChild()
    #        else:
    #            self.manageActions(backwardIdx)
    #        return backwardIdx
    #
    #def expanded(self):
    #    for column in range(self.model().columnCount(Qt.QModelIndex())):
    #        self.resizeColumnToContents(column)

    def onAddRepeat(self):
        index = self.currentIndex()
        if isinstance(self.model(), Qt.QSortFilterProxyModel):
            sourceIndex = self.model().mapToSource(index)
            newSourceIndex = self.model()._insertRow(sourceIndex)
            newIndex = self.model().mapFromSource(newSourceIndex)
        else:
            newIndex = self.model()._insertRow(index)
        self.setCurrentIndex(newIndex)
        self.expandAll()

    def onDelRepeat(self):
        index = self.currentIndex()
        if isinstance(self.model(), Qt.QSortFilterProxyModel):
            index = self.model().mapToSource(index)
        self.model()._removeRow(index)
        self.expandAll()

    def onUpRepeat(self):
        index = self.currentIndex()
        if isinstance(self.model(), Qt.QSortFilterProxyModel):
            sourceIndex = self.model().mapToSource(index)
            newSourceIndex = self.model()._upRow(sourceIndex)
            newIndex = self.model().mapFromSource(newSourceIndex)
        else:
            newIndex = self.model()._upRow(index)
        self.setCurrentIndex(newIndex)
        self.expandAll()

    def onDownRepeat(self):
        index = self.currentIndex()
        if isinstance(self.model(), Qt.QSortFilterProxyModel):
            sourceIndex = self.model().mapToSource(index)
            newSourceIndex = self.model()._downRow(sourceIndex)
            newIndex = self.model().mapFromSource(newSourceIndex)
        else:
            newIndex = self.model()._downRow(index)
        self.setCurrentIndex(newIndex)
        self.expandAll()

class ParamEditorManager(Singleton):

    def init(self):
        self._paths = []
        self._macroEditorsDict = {}

    def paths(self):
        return self._paths

    def setPaths(self, paths):
        self._paths = paths

    def appendPath(self, path):
        self._paths.append()

    def parsePaths(self, pathsString):
        self.setPaths(pathsString.split(":"))

    def browsePaths(self):
        for path in self.paths():
            modulePaths = glob.glob("%s/*.py" % path)
            if not modulePaths:
                continue

            sys.path.insert(0, path)
            for modulePath in modulePaths:
                if modulePath.endswith("__init__.py"):
                    continue
                modulePathAsArray = modulePath.split("/")
                moduleFileName = modulePathAsArray[-1]
                moduleName = moduleFileName[:-3]

                try:
                    module = __import__(moduleName)
                except ImportError:
                    continue
                klass = getattr(module, "CUSTOM_EDITOR", None)
#                for name, klass in inspect.getmembers(module, inspect.isclass):
#                    if issubclass(klass, MacroParametersEditor):
                self._macroEditorsDict[moduleName] = klass

    def getMacroEditor(self, macroName=None, parent=None):
        editorClass = self._macroEditorsDict.get(macroName, None)
        try:
            return editorClass(parent=parent)
        except:
            return None





#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["SardanaBaseElementModel", "SardanaElementTypeModel",
           "SardanaElementPlainModel", "SardanaBaseProxyModel",
           "SardanaTypeProxyModel",
           "SardanaBaseTreeItem", "SardanaRootTreeItem",
           "SardanaElementTreeItem", "SardanaTypeTreeItem",
           "SardanaEnvironmentModel"]

__docformat__ = 'restructuredtext'

try:
    import pygments
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import PythonLexer, PythonTracebackLexer
except:
    pygments = None

from taurus.core.taurusdevice import TaurusDevice
from taurus.external.qt import Qt
from taurus.core.util.enumeration import Enumeration
from taurus.qt.qtcore.model import TaurusBaseTreeItem, TaurusBaseModel, \
    TaurusBaseProxyModel
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE, \
    TAURUS_MODEL_MIME_TYPE

_MOD, _CLS, _FNC, _TNG = ":/python-module.png", ":/class.png", ":/function.png", ":/tango.png"

TYPE_MAP = {
    "ControllerLibrary"    : ("Controller libraries", _MOD, "Controller library",),
    "ControllerClass"      : ("Controller classes", _CLS, "Controller class",),
    "Controller"           : ("Controllers", _TNG, "Controller",),
    "Motor"                : ("Motors", _TNG, "Motor",),
    "PseudoMotor"          : ("Pseudo motors", _TNG, "Pseudo Motor",),
    "CTExpChannel"         : ("Counter/Timers", _TNG, "Counter/Timer experiment channel",),
    "CounterTimer"         : ("Counter/Timers", _TNG, "Counter/Timer experiment channel",),
    "ZeroDExpChannel"      : ("0D channels", _TNG, "0D experiment channel",),
    "OneDExpChannel"       : ("1D channels", _TNG, "1D experiment channel",),
    "TwoDExpChannel"       : ("2D channels", _TNG, "2D experiment channel",),
    "MotorGroup"           : ("Motor groups", _TNG, "Motor group",),
    "MeasurementGroup"     : ("Measurement groups", _TNG, "Measurement group",),
    "CommunicationChannel" : ("Communication channels", _TNG, "Communication channel",),
    "MacroLibrary"         : ("Macro libraries", _MOD, "Macro library",),
    "MacroClass"           : ("Macro classes", _CLS, "Macro class",),
    "Instrument"           : ("Instruments", _TNG, "Instrument",),
    "MacroFunction"        : ("Macro functions", _FNC, "Macro function",),
}

def getElementTypeLabel(t):
    return TYPE_MAP.get(t, (t,))[0]

def getElementTypeIcon(t):
    import taurus.qt.qtgui.resource
    try:
        return taurus.qt.qtgui.resource.getIcon(TYPE_MAP.get(t, (None, _TNG))[1])
    except:
        return None

def getElementTypeSize(t):
    return Qt.QSize(200, 24)

def getElementTypeToolTip(t):
    return TYPE_MAP.get(t, (None, None, 'no information'))[2]


class SardanaBaseTreeItem(TaurusBaseTreeItem):
    """A generic node"""

    def data(self, index):
        """Returns the data of this node for the given index
        
        :return: (object) the data for the given index
        """
        if index.column() > 0:
            return None
        return getElementTypeLabel(self._itemData)

    def role(self):
        """Returns the prefered role for the item.
        This implementation returns taurus.core.taurusbasetypes.TaurusElementType.Unknown
        
        This method should be able to return any kind of python object as long
        as the model that is used is compatible.
        
        :return: the role in form of element type"""
        return 'type'


class SardanaRootTreeItem(SardanaBaseTreeItem):
    pass


class SardanaTypeTreeItem(SardanaBaseTreeItem):
    pass


class SardanaElementTreeItem(SardanaBaseTreeItem):

    def role(self):
        return self.itemData().type

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())
        obj = self.itemData()
        if role == "parent":
            if hasattr(obj, "klass"):
                return obj.klass
            if hasattr(obj, "module"):
                return obj.module
            if hasattr(obj, "controller"):
                return obj.controller
            if hasattr(obj, "parent"):
                return obj.parent
            return None
        return getattr(obj, role)

    def toolTip(self, index):
        if index.column() > 0:
            return self.data(index)
        obj = self.itemData()
        if hasattr(obj, "exc_info") and obj.exc_info is not None:
            html_orig = '<html><head><style type="text/css">{style}' \
                        '</style></head><body>'
            formatter, style = None, ""
            if pygments is not None:
                formatter = HtmlFormatter()
                style = formatter.get_style_defs()
            txt = html_orig.format(style=style)
            if formatter is None:
                txt += "<pre>%s</pre>" % obj.exc_info
            else:
                txt += highlight(obj.exc_info, PythonTracebackLexer(),
                                 formatter)
            txt += "</body></html>"
        else:
            txt = "{0} {1}".format(getElementTypeToolTip(obj.type), obj.name)
        return txt

    def icon(self, index):
        if index.column() > 0:
            return None
        return getElementTypeIcon(self.itemData().type)


class SardanaBaseElementModel(TaurusBaseModel):

    ColumnNames = ["Elements", "Controller/Module/Parent"]
    ColumnRoles = ('Root', 'type', 'name', 'name'), "parent"

    def __init__(self, parent=None, data=None):
        TaurusBaseModel.__init__(self, parent=parent, data=data)
        self.setSelectables(self.ColumnRoles[0])

    def setDataSource(self, data_source):
        old_ds = self.dataSource()
        if old_ds is not None:
            Qt.QObject.disconnect(old_ds, Qt.SIGNAL('elementsChanged'),
                                  self.on_elements_changed)
        if data_source is not None:
            Qt.QObject.connect(data_source, Qt.SIGNAL('elementsChanged'),
                               self.on_elements_changed)
        TaurusBaseModel.setDataSource(self, data_source)

    def on_elements_changed(self):
        self.refresh()

    def createNewRootItem(self):
        return SardanaRootTreeItem(self, self.ColumnNames)

    def roleIcon(self, role):
        return getElementTypeIcon(role)

    def columnIcon(self, column):
        return self.roleIcon(self.role(column))

    def roleToolTip(self, role):
        return getElementTypeToolTip(role)

    def columnToolTip(self, column):
        return self.roleToolTip(self.role(column))

    def roleSize(self, role):
        return getElementTypeSize(role)

    def columnSize(self, column):
        role = self.role(column)
        s = self.roleSize(role)
        return s

    def mimeTypes(self):
        return "text/plain", TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_MODEL_MIME_TYPE

    def mimeData(self, indexes):
        ret = Qt.QMimeData()
        data = []
        for index in indexes:
            if not index.isValid(): continue
            tree_item = index.internalPointer()
            mime_data_item = tree_item.mimeData(index)
            if mime_data_item is None:
                continue
            data.append(mime_data_item)
        ret.setData(TAURUS_MODEL_LIST_MIME_TYPE, "\r\n".join(data))
        ret.setText(", ".join(data))
        if len(data) == 1:
            ret.setData(TAURUS_MODEL_MIME_TYPE, str(data[0]))
        return ret

    def accept(self, element):
        return True

    def setupModelData(self, data):
        dev = self.dataSource()
        if dev is None:
            return
        self.ColumnNames[0] = dev.getSimpleName()
        info = dev.getElementsInfo()
        elements = info.getElements()
        root = self._rootItem
        type_nodes = {}
        parent_elements = {}
        child_elements = set()
        parent_types = "ControllerLibrary", "MacroLibrary", "Controller"
        child_types = "ControllerClass", "MacroClass", "MacroFunction", \
            "Motor", "CounterTimer", "PseudoMotor", "PseudoCounter", \
            "ZeroDExpChannel", "OneDExpChannel", "TwoDExpChannel"

        for element in elements:
            if not self.accept(element):
                continue
            element_type = element.type
            type_item = type_nodes.get(element_type)
            if type_item is None:
                type_item = SardanaTypeTreeItem(self, element_type, root)
                type_nodes[element_type] = type_item
                root.appendChild(type_item)
            element_item = SardanaElementTreeItem(self, element, type_item)
            type_item.appendChild(element_item)
            if element_type in parent_types:
                parent_elements[element.name] = element_item
            elif element_type in child_types:
                child_elements.add(element)

        for element in child_elements:
            try:
                parent_item = parent_elements[element.parent]
            except KeyError:
                self.warning("Error adding %s to parent %s (parent unknown)",
                             element.name, element.parent)
            element_item = SardanaElementTreeItem(self, element, parent_item)
            parent_item.appendChild(element_item)


class SardanaElementTypeModel(SardanaBaseElementModel):
    pass


class SardanaElementPlainModel(SardanaBaseElementModel):

    ColumnNames = "Elements",
    ColumnRoles = ('Root', 'name',),

    def setupModelData(self, data):
        dev = self.dataSource()
        if dev is None:
            return
        info = dev.getElementsInfo()
        elements = info.getElements()
        root = self._rootItem
        skip_types = "ControllerLibrary", "MacroLibrary"

        for element in elements:
            element_type = element.type
            if element_type in skip_types:
                continue
            element_item = SardanaElementTreeItem(self, element, root)
            root.appendChild(element_item)


class SardanaBaseProxyModel(TaurusBaseProxyModel):

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()

        if isinstance(treeItem, SardanaElementTreeItem):
            expr = self.filterRegExp()
            element = treeItem.itemData()
            return self.elementMatches(element, expr)
        return True

    def elementMatches(self, element, expr):
        name = element.name
        if Qt.QString(name).contains(expr):
            return True
        name = element.full_name
        if name is None:
            return False
        return Qt.QString(name).contains(expr)


class SardanaTypeProxyModel(TaurusBaseProxyModel):
    """Sardana filter by element type"""

    def __init__(self, types=None, parent=None):
        TaurusBaseProxyModel.__init__(self, parent=parent)
        if types is None:
            types = ()
        self._types = types

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()

        if isinstance(treeItem, SardanaElementTreeItem):
            return treeItem.itemData().type in self._types
        return False


class EnvironmentTreeItem(TaurusBaseTreeItem):
    """An environment node"""

    def role(self):
        return self.itemData().type

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())
        obj = self.itemData()
        if role == "key":
            return obj[0]
        elif role == "value":
            return obj[1]
        elif role == "datatype":
            return type(obj[1]).__name__

    def toolTip(self, index):
        return "environment '%s'" % self.itemData()[0]

    def icon(self, index):
        if index.column() > 0:
            return None
        return ":/class.png"


class SardanaEnvironmentModel(TaurusBaseModel):

    ColumnNames = ["Environment", "Value", "Data Type"]
    ColumnRoles = ('Root', 'key'), 'value', 'datatype'

    def __init__(self, parent=None, data=None):
        TaurusBaseModel.__init__(self, parent=parent, data=data)
        self.setSelectables(self.ColumnRoles[0])

    def setDataSource(self, data_source):
        old_ds = self.dataSource()
        if old_ds is not None:
            Qt.QObject.disconnect(old_ds, Qt.SIGNAL('environmentChanged'),
                                  self.on_environment_changed)
        if data_source is not None:
            Qt.QObject.connect(data_source, Qt.SIGNAL('environmentChanged'),
                               self.on_environment_changed)
        TaurusBaseModel.setDataSource(self, data_source)

    def on_environment_changed(self):
        self.refresh()

    def createNewRootItem(self):
        return SardanaRootTreeItem(self, self.ColumnNames)

    def roleIcon(self, role):
        return ":/tango.png"
    #    return getElementTypeIcon(role)

    #def columnIcon(self, column):
    #    return self.roleIcon(self.role(column))

    def roleToolTip(self, role):
        cr = self.ColumnRoles
        if role == cr[0][1]:
            return "Environment name"
        elif role == cr[1]:
            return "Environment value"
        elif role == cr[2]:
            return "Environment value data type"

    #def columnToolTip(self, column):
    #    return self.roleToolTip(self.role(column))

    def roleSize(self, role):
        return Qt.QSize(200, 24)

    def columnSize(self, column):
        role = self.role(column)
        s = self.roleSize(role)
        return s

    def mimeTypes(self):
        return "text/plain", TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_MODEL_MIME_TYPE

    def mimeData(self, indexes):
        ret = Qt.QMimeData()
        data = []
        for index in indexes:
            if not index.isValid(): continue
            tree_item = index.internalPointer()
            mime_data_item = tree_item.mimeData(index)
            if mime_data_item is None:
                continue
            data.append(mime_data_item)
        ret.setData(TAURUS_MODEL_LIST_MIME_TYPE, "\r\n".join(data))
        ret.setText(", ".join(data))
        if len(data) == 1:
            ret.setData(TAURUS_MODEL_MIME_TYPE, str(data[0]))
        return ret

    def accept(self, environment):
        return True

    def setupModelData(self, data):
        dev = self.dataSource()
        if dev is None:
            return

        env = dev.getEnvironment()
        root = self._rootItem

        for key, value in env.items():
            if not self.accept(key):
                continue
            env_item = EnvironmentTreeItem(self, (key, value), root)
            root.appendChild(env_item)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
model.py: 
"""

from lxml import etree

from taurus.external.qt import Qt
from sardana.taurus.core.tango.sardana import macro


class MacroSequenceTreeModel(Qt.QAbstractItemModel):

    def __init__(self, parent=None):
        Qt.QAbstractItemModel.__init__(self, parent)
        self.columns = 4
        self.setRoot(macro.SequenceNode())
        self.headers = ["Macro", "Parameters", "Progress", "Pause"]

    def root(self):
        return self._root

    def setRoot(self, root):
        self._root = root
        self.reset()

    def clearSequence(self):
        self.setRoot(macro.SequenceNode())

    def isEmpty(self):
        return len(self.root()) == 0

    def flags(self, index):
        column = index.column()
        node = self.nodeFromIndex(index)
        flags = Qt.Qt.ItemIsEnabled

        if column == 0:
            flags |= Qt.Qt.ItemIsSelectable

        elif column == 1:
            if isinstance(node, macro.SingleParamNode) and \
                not node.type() == "User":
                flags |= Qt.Qt.ItemIsEditable
            else:
                flags |= Qt.Qt.ItemIsSelectable

        elif column == 2:
            flags |= Qt.Qt.ItemIsSelectable

        elif index.column() == 3:
            flags |= (Qt.Qt.ItemIsSelectable | Qt.Qt.ItemIsEditable)

        if isinstance(node, macro.MacroNode):
            flags |= Qt.Qt.ItemIsDragEnabled
            if node.isAllowedHooks():
                flags |= Qt.Qt.ItemIsDropEnabled
        return flags

    def _insertRow(self, parentIndex, node=None, row=-1):
        parentNode = self.nodeFromIndex(parentIndex)

        if row == -1: row = len(parentNode)

        if isinstance(parentNode, macro.RepeatParamNode):
            if node == None: node = parentNode.newRepeat()

        self.beginInsertRows(parentIndex, row, row)
        row = parentNode.insertChild(node, row)
        self.endInsertRows()

        return self.index(row, 0, parentIndex)

    def _removeRow(self, index):
        """This method is used remove macro (pased via index)"""
        node = self.nodeFromIndex(index)
        parentIndex = index.parent()
        parentNode = self.nodeFromIndex(parentIndex)
        row = parentNode.rowOfChild(node)
        self.beginRemoveRows(parentIndex, row, row)
        parentNode.removeChild(node)
        self.endRemoveRows()

    def _upRow(self, index):
        node = self.nodeFromIndex(index)
        parentIndex = index.parent()
        parentNode = self.nodeFromIndex(parentIndex)
        row = parentNode.rowOfChild(node)
        self._removeRow(index)
        newIndex = self._insertRow(parentIndex, node, row - 1)
        if isinstance(parentNode, macro.RepeatParamNode):
            parentNode.arrangeIndexes()
        return newIndex

    def _downRow(self, index):
        node = self.nodeFromIndex(index)
        parentIndex = index.parent()
        parentNode = self.nodeFromIndex(parentIndex)
        row = parentNode.rowOfChild(node)
        self._removeRow(index)
        newIndex = self._insertRow(parentIndex, node, row + 1)
        if isinstance(parentNode, macro.RepeatParamNode):
            parentNode.arrangeIndexes()
        return newIndex

    def _leftRow(self, index):
        """This method is used to move selected macro (pased via index)
        to it's grandparent's hook list. In tree representation it basically move macro to the left"""
        node = self.nodeFromIndex(index)
        parentIndex = index.parent()
        grandParentIndex = parentIndex.parent()
        self._removeRow(index)
        return self._insertRow(grandParentIndex, node)

    def _rightRow(self, index):
        """This method is used to move selected macro (pased via index)
        to it's grandparent's hook list. In tree representation it basically move macro to the left"""
        node = self.nodeFromIndex(index)
        parentIndex = index.parent()
        row = index.row()
        self._removeRow(index)
        newParentIndex = self.index(row, 0, parentIndex)
        return self._insertRow(newParentIndex, node)

    def rowCount(self, parent):
        branchNode = self.nodeFromIndex(parent)
        return len(branchNode)

    def columnCount(self, parent):
        return self.columns

    def data(self, index, role):
        if role == Qt.Qt.DisplayRole:
            node = self.nodeFromIndex(index)
            if index.column() == 0:
                return Qt.QVariant(node.name())
            elif index.column() == 1:
                return Qt.QVariant(node.value())
            elif index.column() == 2:
                if isinstance(node, macro.MacroNode):
                    return Qt.QVariant(node.progress())
        elif role == Qt.Qt.DecorationRole:
            node = self.nodeFromIndex(index)
            if index.column() == 3:
                if isinstance(node, macro.MacroNode):
                    if node.isPause():
                        return Qt.QVariant(Qt.QIcon(":/actions/media-playback-pause.svg"))
        return Qt.QVariant()

    def setData (self, index, value, role=Qt.Qt.EditRole):
        node = self.nodeFromIndex(index)
        if index.column() == 1:
            if isinstance(node, macro.SingleParamNode):
                node.setValue(Qt.from_qvariant(value, str))
                self.emit(Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), index, index)
                while True:
                    index = index.parent()
                    node = self.nodeFromIndex(index)
                    if isinstance(node, macro.MacroNode):
                        self.emit(Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), index, index.sibling(index.row(), self.columnCount(index)))
                        break
        elif index.column() == 2:
            progress = Qt.from_qvariant(value, float)
            node.setProgress(progress)
            self.emit(Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), index, index)
        elif index.column() == 3:
            node.setPause(Qt.from_qvariant(value, bool))
            self.emit(Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), index, index)
        return True


    def headerData(self, section, orientation, role):
        if orientation == Qt.Qt.Horizontal and role == Qt.Qt.DisplayRole:
            return Qt.QVariant(self.headers[section])
        return Qt.QVariant()

    def index(self, row, column, parent):
        assert self.root() is not None
        branchNode = self.nodeFromIndex(parent)
        assert branchNode is not None
        return self.createIndex(row, column, branchNode.child(row))

    def parent(self, child):
        node = self.nodeFromIndex(child)
        if node is None:
            return Qt.QModelIndex()
        parent = node.parent()
        if parent is None:
            return Qt.QModelIndex()
        grandparent = parent.parent()
        if grandparent is None:
            return Qt.QModelIndex()
        row = grandparent.rowOfChild(parent)
        assert row != -1
        return self.createIndex(row, 0, parent)

    def nodeFromIndex(self, index):
        if index.isValid():
            return index.internalPointer()
        else:
            return self.root()

    def toXmlString(self, pretty=False, withId=True):
        xmlSequence = self.root().toXml(withId=withId)
        xmlTree = etree.ElementTree(xmlSequence)
        xmlString = etree.tostring(xmlTree, pretty_print=pretty)
        return xmlString

    def fromXmlString(self, xmlString):
        xmlElement = etree.fromstring(xmlString)
        newRoot = macro.SequenceNode(None)
        newRoot.fromXml(xmlElement)
        self.setRoot(newRoot)
        self.reset()
        return newRoot

    def fromPlainText(self, text):
        newRoot = macro.SequenceNode(None)
        newRoot.fromPlainText(text)
        self.setRoot(newRoot)
        self.reset()
        return newRoot

    def assignIds(self):
        """
        Assigns ids for all macros present in the sequence. If certain macro
        already had an id, it stays without change. A list of all ids is returned
        
        :return: (list)
        """
        parentNode = self.root()
        return self.__assignIds(parentNode)

    def __assignIds(self, parentNode):
        ids = []
        for childNode in parentNode.children():
            if isinstance(childNode, macro.MacroNode):
                id = childNode.assignId()
                ids.append(id)
                ids.extend(self.__assignIds(childNode))
        return ids

    def firstMacroId(self):
        return self.root().child(0).id()

    def lastMacroId(self):
        root = self.root()
        return root.child(len(root.children()) - 1).id()

    def createIdIndexDictionary(self):
        parentIndex = Qt.QModelIndex()
        parentNode = self.root()
        return self.__createIdIndexDictionary(parentIndex, parentNode)

    def __createIdIndexDictionary(self, parentIndex, parentNode):
        d = {}
        for row, child in enumerate(parentNode.children()):
            if isinstance(child, macro.MacroNode):
                index = self.index(row, 0, parentIndex)
                d[child.id()] = index
                d.update(self.__createIdIndexDictionary(index, child))
        return d

#    def supportedDropActions(self):
#        return Qt.Qt.CopyAction | Qt.Qt.MoveAction

#    def mimeTypes(self):
#        types = Qt.QStringList()
#        types.append("text/xml")
#        return types

#    def mimeData(self, indexes):
#        mimeData = Qt.QMimeData()
#        encodedData = Qt.QByteArray()
#        stream = Qt.QDataStream(encodedData, Qt.QIODevice.WriteOnly)
#        doc = xml.dom.minidom.Document()
#        for i,index in enumerate(indexes):
#            if i % 2:
#                continue
#            text = self.nodeFromIndex(index).toXml(doc).toxml()
#            stream.writeString(text)
#
#        mimeData.setData("text/xml", encodedData)
#        return mimeData
#
#    def dropMimeData(self, data, action, row, column, parent):
#        if action == Qt.Qt.IgnoreAction:
#            return True
#        if not data.hasFormat("text/xml"):
#            return False
#
#        encodedData = data.data("text/xml")
#        stream = Qt.QDataStream(encodedData, Qt.QIODevice.ReadOnly)
#        newItems = Qt.QStringList()
#        rows = 0
#
#        while(not stream.atEnd()):
#            text = stream.readString()
#            newItems.append(text)
#            rows += 1
#
#        sequence = self.nodeFromIndex(parent)
#
#        for text in newItems:
#            macroNode = macro.MacroNode()
#            macroNode.fromDoc(xml.dom.minidom.parseString(text))
#            self.insertMacro(sequence, macroNode, row, False)
#            macros = [macro.name() for macro in macroNode.allMacros()]
#            if action == Qt.Qt.CopyAction:
#                self.emit(Qt.SIGNAL("macrosAdded"), macros, macroNode.allMotors())
#            self.emit(Qt.SIGNAL("dataChanged"))
#        return True


class MacroSequenceProxyModel(Qt.QSortFilterProxyModel):

    def __init__(self, parent=None):
        Qt.QSortFilterProxyModel.__init__(self, parent)
        self.setDynamicSortFilter(True)
        self.headers = ["Macro", "Parameters", "Progress", "Pause"]
        self.columns = 4

    def __getattr__(self, name):
        return getattr(self.sourceModel(), name)

    def nodeFromIndex(self, index):
        sourceIndex = self.mapToSource(index)
        node = self.sourceModel().nodeFromIndex(sourceIndex)
        return node

    def createIdIndexDictionary(self):
        d = self.sourceModel().createIdIndexDictionary()
        for id, sourceIndex in d.iteritems():
            proxyIndex = self.mapFromSource(sourceIndex)
            d[id] = Qt.QPersistentModelIndex(proxyIndex)
        return d

    def filterAcceptsRow(self, row, parentIndex):
        child = self.sourceModel().index(row, 0, parentIndex)
        node = self.sourceModel().nodeFromIndex(child)
        return isinstance(node, macro.MacroNode)

class MacroParametersProxyModel(Qt.QSortFilterProxyModel):

    def __init__(self, parent=None):
        Qt.QSortFilterProxyModel.__init__(self, parent)
        self.columns = 2
        self.headers = ["Parameter", "Value", "", "", "", ""]
        self._macroIndex = None

    def __getattr__(self, name):
        return getattr(self.sourceModel(), name)

    def headerData(self, section, orientation, role):
        if orientation == Qt.Qt.Horizontal and role == Qt.Qt.DisplayRole:
            return Qt.QVariant(self.headers[section])
        return Qt.QVariant()

    def nodeFromIndex(self, index):
        sourceIndex = self.mapToSource(index)
        node = self.sourceModel().nodeFromIndex(sourceIndex)
        return node

    def setMacroIndex(self, macroIndex):
        self._macroIndex = macroIndex

    def macroIndex(self):
        return self._macroIndex

    def columnCount(self, parent):
        return self.columns

    def filterAcceptsRow(self, row, parentIndex):
        if self.macroIndex() == None:
            return False
        if self.macroIndex() == parentIndex:
            child = self.sourceModel().index(row, 0, parentIndex)
            node = self.sourceModel().nodeFromIndex(child)
            if not isinstance(node, macro.ParamNode):
                return False
        return True






#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the class for an
abstract action over a set of pool elements"""

__all__ = ["PoolActionItem", "OperationInfo", "ActionContext", "PoolAction",
           "get_thread_pool"]

__docformat__ = 'restructuredtext'

import sys
import weakref
import traceback
import threading

from taurus.core.util.log import Logger

from sardana import State
from sardana.sardanathreadpool import get_thread_pool
from sardana.pool.poolobject import PoolObject


class PoolActionItem(object):
    """The base class for an atomic action item"""

    def __init__(self, element):
        self._element = weakref.ref(element)

    def get_element(self):
        """Returns the element associated with this item"""
        return self._element()

    def set_element(self, element):
        """Sets the element for this item"""
        self._element = weakref.ref(element)

    element = property(get_element)


class OperationInfo(object):
    """Stores synchronization data for a certain operation"""

    def __init__(self):
        """Constructor"""
        self.state_count = 0
        self.state_lock = threading.Lock()
        self.state_event_lock = threading.Lock()
        self.state_event = threading.Event()

    def init(self, count):
        """Initializes this operation with a certain count"""
        self.state_count = count
        self.state_event.clear()
        if count == 0:
            self.state_event.set()

    def wait(self, timeout=None):
        """waits for the operation to finish"""
        return self.state_event.wait(timeout)

    def finish_one(self):
        """Notifies this operation that one step was finished"""
        with self.state_event_lock:
            self.state_count = self.state_count - 1
            if self.state_count < 1:
                self.state_count = 0
                self.state_event.set()

    def acquire(self):
        """Acquires this operation lock"""
        self.state_lock.acquire()

    def release(self):
        """Releases this operation lock"""
        self.state_lock.release()

    def __enter__(self):
        return self.acquire()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return self.release()


class BaseOperationContext(object):
    """Stores operation context"""

    def __init__(self, pool_action):
        self._pool_action = pool_action

    def enter(self):
        """Enters operation context"""
        pool_action = self._pool_action
        for element in pool_action.get_elements():
            element.lock()
            element.set_operation(pool_action)
        for ctrl in pool_action.get_pool_controller_list():
            ctrl.lock()

    def exit(self):
        """Leaves operation context"""
        pool_action = self._pool_action
        for element in reversed(pool_action.get_elements()):
            element.clear_operation()
            element.unlock()
        for ctrl in reversed(pool_action.get_pool_controller_list()):
            ctrl.unlock()
        pool_action.finish_action()
        return False

    def __enter__(self):
        return self.enter()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return self.exit()


class OperationContext(BaseOperationContext):
    """Concrete operation context"""

    def enter(self):
        pool_action = self._pool_action
        for element in pool_action.get_elements():
            element.set_operation(pool_action)

    def exit(self):
        pool_action = self._pool_action
        for element in reversed(pool_action.get_elements()):
            element.clear_operation()
        pool_action.finish_action()
        return False

    def __enter__(self):
        return self.enter()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return self.exit()


class ActionContext(object):
    """Stores an atomic action context"""

    def __init__(self, pool_action):
        self._pool_action = pool_action

    def enter(self):
        """Enters operation"""
        pool_action = self._pool_action
        for element in pool_action.get_elements():
            element.lock()
        for ctrl in pool_action.get_pool_controller_list():
            ctrl.lock()

    def exit(self):
        """Leaves operation"""
        pool_action = self._pool_action
        for element in reversed(pool_action.get_elements()):
            element.unlock()
        for ctrl in reversed(pool_action.get_pool_controller_list()):
            ctrl.unlock()
        return False

    def __enter__(self):
        return self.enter()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return self.exit()


class PoolAction(Logger):
    """A generic class to handle any type of operation (like motion or
    acquisition)"""

    def __init__(self, main_element, name="GlobalAction"):
        Logger.__init__(self, name)
        self._action_run_lock = threading.Lock()
        self._main_element = weakref.ref(main_element)
        self._aborted = False
        self._stopped = False
        self._elements = []
        self._pool_ctrl_dict = {}
        self._pool_ctrl_list = []
        self._finish_hook = None
        self._running = False
        self._state_info = OperationInfo()
        self._value_info = OperationInfo()

    def get_main_element(self):
        """Returns the main element for this action

        :return: sardana.pool.poolelement.PoolElement"""
        return self._main_element()

    main_element = property(get_main_element)

    def get_pool(self):
        """Returns the pool object for thi action

        :return: sardana.pool.pool.Pool"""
        return self.main_element.pool

    pool = property(get_pool)

    def clear_elements(self):
        """Clears all elements from this action"""
        self._elements = []
        self._pool_ctrl_dict = {}
        self._pool_ctrl_list = []

    def add_element(self, element):
        """Adds a new element to this action.

        :param element: the new element to be added
        :type element: sardana.pool.poolelement.PoolElement"""
        ctrl = element.controller
        ctrl_items = self._pool_ctrl_dict.get(ctrl)
        if ctrl_items is None:
            ctrl_items = []
            self._pool_ctrl_dict[ctrl] = ctrl_items
            self._pool_ctrl_list.append(ctrl)
            self._pool_ctrl_list.sort(key=PoolObject.get_id)

        self._elements.append(element)
        ctrl_items.append(element)
        # make sure elements are ordered by ID so that a multiple lock always
        # locks and unlocks in the same order
        self._elements.sort(key=PoolObject.get_id)
        ctrl_items.sort(key=PoolObject.get_id)

    def remove_element(self, element):
        """Removes an element from this action. If the element is not part of
        this action, a ValueError is raised.

        :param element: the new element to be removed
        :type element: sardana.pool.poolelement.PoolElement

        :raises: ValueError"""
        ctrl = element.controller
        #element = weakref.ref(element)
        try:
            idx = self._elements.index(element)
        except ValueError:
            raise ValueError("action doesn't contain %s" % element.name)
        del self._elements[idx]
        ctrl_items = self._pool_ctrl_dict[ctrl]
        del ctrl_items[ctrl_items.index(element)]
        if not len(ctrl_items):
            del self._pool_ctrl_dict[ctrl]
            del self._pool_ctrl_list[self._pool_ctrl_list.index(ctrl)]

    def get_elements(self, copy_of=False):
        """Returns a sequence of all elements involved in this action.

        :param copy_of: If False (default) the internal container of elements is
                        returned. If True, a copy of the internal container is
                        returned instead
        :type copy_of: bool
        :return: a sequence of all elements involved in this action.
        :rtype: seq<sardana.pool.poolelement.PoolElement>"""
        elements = self._elements
        if copy_of:
            elements = tuple(elements)
        return elements

    def get_pool_controller_list(self):
        """Returns a list of all controller elements involved in this action.

        :return: a list of all controller elements involved in this action.
        :rtype: list<sardana.pool.poolelement.PoolController>"""
        return self._pool_ctrl_list

    def get_pool_controllers(self):
        """Returns a dict of all controller elements involved in this action.

        :return: a dict of all controller elements involved in this action.
        :rtype: dict<sardana.pool.poolelement.PoolController,
                     seq<sardana.pool.poolelement.PoolElement>>"""
        return self._pool_ctrl_dict

    def _is_in_action(self, state):
        """Determines if the given state is a busy state (Moving or Running) or
        not.

        :return: True if state is a busy state or False otherwise
        :rtype: bool"""
        return state == State.Moving or state == State.Running

    def is_running(self):
        """Determines if this action is running or not

        :return: True if action is running or False otherwise
        :rtype: bool"""
        return self._running

    def run(self, *args, **kwargs):
        """Runs this action"""

        self._running = True
        synch = kwargs.pop("synch", False)

        if synch:
            try:
                with OperationContext(self) as context:
                    self.start_action(*args, **kwargs)
                    self.action_loop()
            finally:
                self._running = False
        else:
            context = OperationContext(self)
            context.enter()
            try:
                self.start_action(*args, **kwargs)
            except:
                context.exit()
                self._running = False
                raise
            get_thread_pool().add(self._asynch_action_loop, None, context)

    def start_action(self, *args, **kwargs):
        """Start procedure for this action. Default implementation raises
        NotImplementedError

        :raises: NotImplementedError"""
        raise NotImplementedError("start_action must be implemented in "
                                  "subclass")

    def set_finish_hook(self, hook):
        """Attaches/Detaches a finish hook

        :param hook: a callable object or None
        :type hook: callable or None"""
        self._finish_hook = hook

    def finish_action(self):
        """Finishes the action execution. If a finish hook is defined it safely
        executes it. Otherwise nothing happens"""
        hook = self._finish_hook
        if hook is None:
            return
        try:
            hook()
        except:
            self.warning("Exception running function finish hook", exc_info=1)

    def stop_action(self, *args, **kwargs):
        """Stop procedure for this action."""
        self._stopped = True
        for pool_ctrl, elements in self._pool_ctrl_dict.items():
            pool_ctrl.stop_elements(elements)

    def abort_action(self, *args, **kwargs):
        """Aborts procedure for this action"""
        self._aborted = True
        for pool_ctrl, elements in self._pool_ctrl_dict.items():
            pool_ctrl.abort_elements(elements)

    def emergency_break(self):
        """Tries to execute a stop. If it fails try an abort"""
        self._stopped = True
        for pool_ctrl, elements in self._pool_ctrl_dict.items():
            pool_ctrl.emergency_break(elements)

    def was_stopped(self):
        """Determines if the action has been stopped from outside

        :return: True if action has been stopped from outside or False otherwise
        :rtype: bool"""
        return self._stopped

    def was_aborted(self):
        """Determines if the action has been aborted from outside

        :return: True if action has been aborted from outside or False otherwise
        :rtype: bool"""
        return self._aborted

    def was_action_interrupted(self):
        """Determines if the action has been interruped from outside (either
        from an abort or a stop).

        :return: True if action has been interruped from outside or False
                 otherwise
        :rtype: bool"""
        return self.was_aborted() or self.was_stopped()

    def _asynch_action_loop(self, context):
        """Internal method. Asynchronous action loop"""
        try:
            self.action_loop()
        finally:
            context.exit()
            self._running = False

    def action_loop(self):
        """Action loop for this action. Default implementation raises
        NotImplementedError

        :raises: NotImplementedError"""
        raise NotImplementedError("action_loop must be implemented in subclass")

    def read_state_info(self, ret=None, serial=False):
        """Reads state information of all elements involved in this action

        :param ret: output map parameter that should be filled with state
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW state requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing state information per element
        :rtype: dict<sardana.pool.poolelement.PoolElement, stateinfo>"""
        with ActionContext(self):
            return self.raw_read_state_info(ret=ret, serial=serial)

    def raw_read_state_info(self, ret=None, serial=False):
        """**Unsafe**. Reads state information of all elements involved in this
        action

        :param ret: output map parameter that should be filled with state
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW state requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing state information per element
        :rtype: dict<sardana.pool.poolelement.PoolElement, stateinfo>"""
        if ret is None:
            ret = {}
        read = self._raw_read_state_info_concurrent
        if serial:
            read = self._raw_read_state_info_serial
        state_info = self._state_info

        with state_info:
            state_info.init(len(self._pool_ctrl_dict))
            read(ret)
            state_info.wait()
        return ret

    def _raw_read_state_info_serial(self, ret):
        """Internal method. Read state in a serial mode"""
        for pool_ctrl in self._pool_ctrl_dict:
            self._raw_read_ctrl_state_info(ret, pool_ctrl)
        return ret

    def _raw_read_state_info_concurrent(self, ret):
        """Internal method. Read state in a concurrent mode"""
        th_pool = get_thread_pool()
        for pool_ctrl in self._pool_ctrl_dict:
            th_pool.add(self._raw_read_ctrl_state_info, None, ret, pool_ctrl)
        return ret

    def _get_ctrl_error_state_info(self, pool_ctrl):
        """Internal method. Returns the controller error in form of a
        tuple<sardana.State, str>"""
        exc_t, exc_v, trb = sys.exc_info()
        if exc_t is None:
            if pool_ctrl.is_online():
                return State.Fault, "Unknown controller error"
        else:
            if pool_ctrl.is_online():
                err_msg = "".join(traceback.format_exception(exc_t, exc_v, trb))
                return State.Fault, "Unexpected controller error:\n" + err_msg
        return State.Fault, pool_ctrl.get_ctrl_error_str()

    def _raw_read_ctrl_state_info(self, ret, pool_ctrl):
        """Internal method. Read controller information and store it in ret
        parameter"""
        try:
            axes = [elem.axis for elem in self._pool_ctrl_dict[pool_ctrl]]
            state_infos, error = pool_ctrl.raw_read_axis_states(axes)
            if error:
                pool_ctrl.warning("Read state error")
                for elem, (state_info, exc_info) in state_infos.items():
                    if exc_info is not None:
                        pool_ctrl.debug("Axis %s error details:", elem.axis,
                                        exc_info=exc_info)
            ret.update(state_infos)
        except:
            self.error("Something wrong happend: Error should have been caught"
                       "by ctrl.read_axis_states")
            self.debug("Details: ", exc_info=1)
            state_info = self._get_ctrl_error_state_info(pool_ctrl)
            for elem in self._pool_ctrl_dict[pool_ctrl]:
                ret[elem] = state_info
        finally:
            self._state_info.finish_one()

    def get_read_value_ctrls(self):
        return self._pool_ctrl_dict

    def read_value(self, ret=None, serial=False):
        """Reads value information of all elements involved in this action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement`,
                     (value object, Exception or None)>"""
        with ActionContext(self):
            return self.raw_read_value(ret=ret, serial=serial)

    def raw_read_value(self, ret=None, serial=False):
        """**Unsafe**. Reads value information of all elements involved in this
        action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement,
                :class:`sardana.sardanavalue.SardanaValue` >"""

        if ret is None:
            ret = {}

        read = self._raw_read_value_concurrent
        if serial:
            read = self._raw_read_value_serial

        value_info = self._value_info

        with value_info:
            value_info.init(len(self.get_read_value_ctrls()))
            read(ret)
            value_info.wait()
        return ret

    def _raw_read_value_serial(self, ret):
        """Internal method. Read value in a serial mode"""
        for pool_ctrl in self.get_read_value_ctrls():
            self._raw_read_ctrl_value(ret, pool_ctrl)
        return ret

    def _raw_read_value_concurrent(self, ret):
        """Internal method. Read value in a concurrent mode"""
        th_pool = get_thread_pool()
        for pool_ctrl in self.get_read_value_ctrls():
            th_pool.add(self._raw_read_ctrl_value, None, ret, pool_ctrl)
        return ret

    def _raw_read_ctrl_value(self, ret, pool_ctrl):
        """Internal method. Read controller value information and store it in
        ret parameter"""
        try:
            axes = [elem.axis for elem in self._pool_ctrl_dict[pool_ctrl]]
            value_infos = pool_ctrl.raw_read_axis_values(axes)
            ret.update(value_infos)
        finally:
            self._value_info.finish_one()

    def get_read_value_loop_ctrls(self):
        return self._pool_ctrl_dict

    def read_value_loop(self, ret=None, serial=False):
        """Reads value information of all elements involved in this action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement`,
                     (value object, Exception or None)>"""
        with ActionContext(self):
            return self.raw_read_value_loop(ret=ret, serial=serial)

    def raw_read_value_loop(self, ret=None, serial=False):
        """**Unsafe**. Reads value information of all elements involved in this
        action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement,
                :class:`sardana.sardanavalue.SardanaValue` >"""

        if ret is None:
            ret = {}

        read = self._raw_read_value_concurrent_loop
        if serial:
            read = self._raw_read_value_serial_loop

        value_info = self._value_info

        with value_info:
            value_info.init(len(self.get_read_value_loop_ctrls()))
            read(ret)
            value_info.wait()
        return ret

    def _raw_read_value_serial_loop(self, ret):
        """Internal method. Read value in a serial mode"""
        for pool_ctrl in self.get_read_value_loop_ctrls():
            self._raw_read_ctrl_value(ret, pool_ctrl)
        return ret

    def _raw_read_value_concurrent_loop(self, ret):
        """Internal method. Read value in a concurrent mode"""
        th_pool = get_thread_pool()
        for pool_ctrl in self.get_read_value_loop_ctrls():
            th_pool.add(self._raw_read_ctrl_value, None, ret, pool_ctrl)
        return ret



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module provides widget for configuring the data acquisition and display of an experiment"""

__all__ = ["ExpDescriptionEditor"]

from taurus.external.qt import Qt
import copy
import taurus
import taurus.core
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui import resource

from sardana.taurus.qt.qtcore.tango.sardana.model import SardanaBaseProxyModel, SardanaTypeTreeItem
from taurus.qt.qtgui.util.ui import UILoadable

## Using a plain model and filtering and checking 'Acquirable' in item.itemData().interfaces is more elegant, but things don't get properly sorted...
#from taurus.qt.qtcore.tango.sardana.model import SardanaElementPlainModel

class SardanaAcquirableProxyModel(SardanaBaseProxyModel):
#    ALLOWED_TYPES = 'Acquirable'
#
#    def filterAcceptsRow(self, sourceRow, sourceParent):
#        sourceModel = self.sourceModel()
#        idx = sourceModel.index(sourceRow, 0, sourceParent)
#        item = idx.internalPointer()
#        return 'Acquirable' in item.itemData().interfaces

#    ALLOWED_TYPES = ['Motor', 'CTExpChannel', 'ZeroDExpChannel', 'OneDExpChannel',
#                     'TwoDExpChannel', 'ComChannel', 'IORegister', 'PseudoMotor',
#                     'PseudoCounter']

    from sardana.sardanadefs import ElementType, TYPE_ACQUIRABLE_ELEMENTS
    ALLOWED_TYPES = [ElementType[t] for t in TYPE_ACQUIRABLE_ELEMENTS]

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()
        if isinstance(treeItem, SardanaTypeTreeItem):
            return treeItem.itemData() in self.ALLOWED_TYPES
        return True


@UILoadable(with_ui='ui')
class ExpDescriptionEditor(Qt.QWidget, TaurusBaseWidget):
    '''
    A widget for editing the configuration of a experiment (measurement groups,
    plot and storage parameters, etc).
    
    It receives a Sardana Door name as its model and gets/sets the configuration
    using the `ExperimentConfiguration` environmental variable for that Door.
    '''
    def __init__(self, parent=None, door=None, plotsButton=True):
        Qt.QWidget.__init__(self, parent)
        TaurusBaseWidget.__init__(self, 'ExpDescriptionEditor')
        self.loadUi()
        self.ui.buttonBox.setStandardButtons(Qt.QDialogButtonBox.Reset | Qt.QDialogButtonBox.Apply)
        newperspectivesDict = copy.deepcopy(self.ui.sardanaElementTree.KnownPerspectives)
        #newperspectivesDict[self.ui.sardanaElementTree.DftPerspective]['model'] = [SardanaAcquirableProxyModel, SardanaElementPlainModel]
        newperspectivesDict[self.ui.sardanaElementTree.DftPerspective]['model'][0] = SardanaAcquirableProxyModel
        self.ui.sardanaElementTree.KnownPerspectives = newperspectivesDict  #assign a copy because if just a key of this class memberwas modified, all instances of this class would be affected
        self.ui.sardanaElementTree._setPerspective(self.ui.sardanaElementTree.DftPerspective)

        self._localConfig = None
        self._originalConfiguration = None
        self._dirty = False
        self._dirtyMntGrps = set()

        self.connect(self.ui.activeMntGrpCB, Qt.SIGNAL('activated (QString)'), self.changeActiveMntGrp)
        self.connect(self.ui.createMntGrpBT, Qt.SIGNAL('clicked ()'), self.createMntGrp)
        self.connect(self.ui.deleteMntGrpBT, Qt.SIGNAL('clicked ()'), self.deleteMntGrp)
        self.connect(self.ui.compressionCB, Qt.SIGNAL('currentIndexChanged (int)'), self.onCompressionCBChanged)
        self.connect(self.ui.pathLE, Qt.SIGNAL('textEdited (QString)'), self.onPathLEEdited)
        self.connect(self.ui.filenameLE, Qt.SIGNAL('textEdited (QString)'), self.onFilenameLEEdited)
        self.connect(self.ui.channelEditor.getQModel(), Qt.SIGNAL('dataChanged (QModelIndex, QModelIndex)'), self._updateButtonBox)
        self.connect(self.ui.channelEditor.getQModel(), Qt.SIGNAL('modelReset ()'), self._updateButtonBox)
        self.connect(self.ui.preScanList, Qt.SIGNAL('dataChanged'), self.onPreScanSnapshotChanged)
        self.connect(self.ui.choosePathBT, Qt.SIGNAL('clicked ()'), self.onChooseScanDirButtonClicked)
        
        self.__plotManager = None
        icon = resource.getIcon(":/actions/view.svg")
        self.togglePlotsAction = Qt.QAction(icon, "Show/Hide plots", self)
        self.togglePlotsAction.setCheckable(True)
        self.togglePlotsAction.setChecked(False)
        self.togglePlotsAction.setEnabled(plotsButton)
        self.addAction(self.togglePlotsAction)
        self.connect(self.togglePlotsAction, Qt.SIGNAL("toggled(bool)"), 
                     self.onPlotsButtonToggled)
        self.ui.plotsButton.setDefaultAction(self.togglePlotsAction)
        
        if door is not None:
            self.setModel(door)
        self.connect(self.ui.buttonBox, Qt.SIGNAL("clicked(QAbstractButton *)"), self.onDialogButtonClicked)

        #Taurus Configuration properties and delegates
        self.registerConfigDelegate(self.ui.channelEditor)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return taurus.core.taurusdevice.TaurusDevice

    def onChooseScanDirButtonClicked(self):
        ret = Qt.QFileDialog.getExistingDirectory (self, 'Choose directory for saving files', self.ui.pathLE.text())
        if ret:
            self.ui.pathLE.setText(ret)
            self.ui.pathLE.emit(Qt.SIGNAL('textEdited (QString)'), ret)

    def onDialogButtonClicked(self, button):
        role = self.ui.buttonBox.buttonRole(button)
        if role == Qt.QDialogButtonBox.ApplyRole:
            self.writeExperimentConfiguration(ask=False)
        elif role == Qt.QDialogButtonBox.ResetRole:
            self._reloadConf()

    def closeEvent(self, event):
        '''This event handler receives widget close events'''
        if self.isDataChanged():
            self.writeExperimentConfiguration(ask=True)
        Qt.QWidget.closeEvent(self, event)

    def setModel(self, model):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        TaurusBaseWidget.setModel(self, model)
        self._reloadConf(force=True)
        #set the model of some child widgets
        door = self.getModelObj()
        if door is None: return
        tghost = taurus.Database().getNormalName()  #@todo: get the tghost from the door model instead
        msname = door.macro_server.getFullName()
        self.ui.taurusModelTree.setModel(tghost)
        self.ui.sardanaElementTree.setModel(msname)

    def _reloadConf(self, force=False):
        if not force and self.isDataChanged():
            op = Qt.QMessageBox.question(self, "Reload info from door",
                "If you reload, all current experiment configuration changes will be lost. Reload?",
                Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
            if op != Qt.QMessageBox.Yes:
                return
        door = self.getModelObj()
        if door is None: return
        conf = door.getExperimentConfiguration()
        self._originalConfiguration = copy.deepcopy(conf)
        self.setLocalConfig(conf)
        self._setDirty(False)
        self._dirtyMntGrps = set()
        #set a list of available channels
        avail_channels = {}
        for ch_info in door.macro_server.getExpChannelElements().values():
            avail_channels[ch_info.full_name] = ch_info.getData()
        self.ui.channelEditor.getQModel().setAvailableChannels(avail_channels)

    def _setDirty(self, dirty):
        self._dirty = dirty
        self._updateButtonBox()

    def isDataChanged(self):
        """Tells if the local data has been modified since it was last refreshed
        
        :return: (bool) True if he local data has been modified since it was last refreshed
        """
        return bool(self._dirty or self.ui.channelEditor.getQModel().isDataChanged() or self._dirtyMntGrps)

    def _updateButtonBox(self, *args, **kwargs):
        self.ui.buttonBox.setEnabled(self.isDataChanged())

    def getLocalConfig(self):
        return self._localConfig

    def setLocalConfig(self, conf):
        '''gets a ExpDescription dictionary and sets up the widget'''

        self._localConfig = conf

        #set the Channel Editor
        activeMntGrpName = self._localConfig['ActiveMntGrp'] or ''
        if activeMntGrpName in self._localConfig['MntGrpConfigs']:
            mgconfig = self._localConfig['MntGrpConfigs'][activeMntGrpName]
            self.ui.channelEditor.getQModel().setDataSource(mgconfig)

        #set the measurement group ComboBox
        self.ui.activeMntGrpCB.clear()
        self.ui.activeMntGrpCB.addItems(sorted(self._localConfig['MntGrpConfigs'].keys()))
        idx = self.ui.activeMntGrpCB.findText(activeMntGrpName)
        self.ui.activeMntGrpCB.setCurrentIndex(idx)

        #set the system snapshot list
        psl = self._localConfig.get('PreScanSnapshot')  #I get it before clearing because clear() changes the _localConfig
        self.ui.preScanList.clear()
        self.ui.preScanList.addModels(psl)

        #other settings
        self.ui.filenameLE.setText(", ".join(self._localConfig['ScanFile']))
        self.ui.pathLE.setText(self._localConfig['ScanDir'] or '')
        self.ui.compressionCB.setCurrentIndex(self._localConfig['DataCompressionRank'] + 1)

    def writeExperimentConfiguration(self, ask=True):
        '''sends the current local configuration to the door
        
        :param ask: (bool) If True (default) prompts the user before saving.
        '''

        if ask:
            op = Qt.QMessageBox.question(self, "Save configuration?",
                                        'Do you want to save the current configuration?\n(if not, any changes will be lost)',
                                        Qt.QMessageBox.Yes | Qt.QMessageBox.No)
            if op != Qt.QMessageBox.Yes:
                return False

        conf = self.getLocalConfig()

        #make sure that no empty measurement groups are written
        for mgname, mgconfig in conf.get('MntGrpConfigs', {}).items():
            if mgconfig is not None and not mgconfig.get('controllers'):
                Qt.QMessageBox.information(self, "Empty Measurement group",
                "The measurement group '%s' is empty. Fill it (or delete it) before applying" % mgname,
                Qt.QMessageBox.Ok)
                self.changeActiveMntGrp(mgname)
                return False

        #check if the currently displayed mntgrp is changed
        if self.ui.channelEditor.getQModel().isDataChanged():
            self._dirtyMntGrps.add(self._localConfig['ActiveMntGrp'])

        door = self.getModelObj()
        door.setExperimentConfiguration(conf, mnt_grps=self._dirtyMntGrps)
        self._originalConfiguration = copy.deepcopy(conf)
        self._dirtyMntGrps = set()
        self.ui.channelEditor.getQModel().setDataChanged(False)
        self._setDirty(False)
        self.emit(Qt.SIGNAL('experimentConfigurationChanged'), copy.deepcopy(conf))
        return True

    def changeActiveMntGrp(self, activeMntGrpName):
        activeMntGrpName = str(activeMntGrpName)
        if self._localConfig is None:
            return
        if activeMntGrpName == self._localConfig['ActiveMntGrp']:
            return  #nothing changed
        if activeMntGrpName not in self._localConfig['MntGrpConfigs']:
            raise KeyError('Unknown measurement group "%s"' % activeMntGrpName)

        #add the previous measurement group to the list of "dirty" groups if something was changed
        if self.ui.channelEditor.getQModel().isDataChanged():
            self._dirtyMntGrps.add(self._localConfig['ActiveMntGrp'])

        self._localConfig['ActiveMntGrp'] = activeMntGrpName

        i = self.ui.activeMntGrpCB.findText(activeMntGrpName, Qt.Qt.MatchExactly)
        self.ui.activeMntGrpCB.setCurrentIndex(i)
        mgconfig = self._localConfig['MntGrpConfigs'][activeMntGrpName]
        self.ui.channelEditor.getQModel().setDataSource(mgconfig)
        self._setDirty(True)

    def createMntGrp(self):
        '''creates a new Measurement Group'''

        if self._localConfig is None:
            return

        mntGrpName, ok = Qt.QInputDialog.getText(self, "New Measurement Group",
                                                 "Enter a name for the new measurement Group")
        if not ok: return
        mntGrpName = str(mntGrpName)

        #check that the given name is not an existing pool element
        ms = self.getModelObj().macro_server
        poolElementNames = [v.name for v in ms.getElementsWithInterface("PoolElement").values()]
        while mntGrpName in poolElementNames:
            Qt.QMessageBox.warning(self, "Cannot create Measurement group",
                "The name '%s' already is used for another pool element. Please Choose a different one." % mntGrpName,
                Qt.QMessageBox.Ok)
            mntGrpName, ok = Qt.QInputDialog.getText(self, "New Measurement Group",
                                                     "Enter a name for the new measurement Group",
                                                     Qt.QLineEdit.Normal,
                                                     mntGrpName)
            if not ok: return
            mntGrpName = str(mntGrpName)

        #check that the measurement group is not already in the localConfig
        if mntGrpName in self._localConfig['MntGrpConfigs']:
            Qt.QMessageBox.warning(self, "%s already exists" % mntGrpName,
                'A measurement group named "%s" already exists. A new one will not be created' % mntGrpName)
            return

        #add an empty configuration dictionary to the local config
        mgconfig = {'label': mntGrpName, 'controllers':{} }
        self._localConfig['MntGrpConfigs'][mntGrpName] = mgconfig
        #add the new measurement group to the list of "dirty" groups
        self._dirtyMntGrps.add(mntGrpName)
        #add the name to the combobox
        self.ui.activeMntGrpCB.addItem(mntGrpName)
        #make it the Active MntGrp
        self.changeActiveMntGrp(mntGrpName)

    def deleteMntGrp(self):
        '''creates a new Measurement Group'''
        activeMntGrpName = str(self.ui.activeMntGrpCB.currentText())
        op = Qt.QMessageBox.question(self, "Delete Measurement Group",
                "Remove the measurement group '%s'?" % activeMntGrpName,
                Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
        if op != Qt.QMessageBox.Yes:
            return
        currentIndex = self.ui.activeMntGrpCB.currentIndex()
        if self._localConfig is None:
            return
        if activeMntGrpName not in self._localConfig['MntGrpConfigs']:
            raise KeyError('Unknown measurement group "%s"' % activeMntGrpName)

        #add the current measurement group to the list of "dirty" groups
        self._dirtyMntGrps.add(activeMntGrpName)

        self._localConfig['MntGrpConfigs'][activeMntGrpName] = None
        self.ui.activeMntGrpCB.setCurrentIndex(-1)
        self.ui.activeMntGrpCB.removeItem(currentIndex)
        self.ui.channelEditor.getQModel().setDataSource({})
        self._setDirty(True)

    def onCompressionCBChanged(self, idx):
        if self._localConfig is None: return
        self._localConfig['DataCompressionRank'] = idx - 1
        self._setDirty(True)

    def onPathLEEdited(self, text):
        self._localConfig['ScanDir'] = str(text)
        self._setDirty(True)

    def onFilenameLEEdited(self, text):
        self._localConfig['ScanFile'] = [v.strip() for v in str(text).split(',')]
        self._setDirty(True)

    def onPreScanSnapshotChanged(self, items):
        door = self.getModelObj()
        ms = door.macro_server
        preScanList = []
        for e in items:
            nfo = ms.getElementInfo(e.src)
            if nfo is None:
                preScanList.append((e.src, e.display))
            else:
                preScanList.append((nfo.full_name, nfo.name))
        self._localConfig['PreScanSnapshot'] = preScanList
        self._setDirty(True)
        
    def onPlotsButtonToggled(self, checked):
        if checked:
            from taurus.qt.qtgui.taurusgui.macrolistener import \
                 DynamicPlotManager
            self.__plotManager = DynamicPlotManager(self)
            self.__plotManager.setModel(self.getModelName())
            self.connect(self, Qt.SIGNAL('experimentConfigurationChanged'), 
                         self.__plotManager.onExpConfChanged)
        else:
            self.disconnect(self, Qt.SIGNAL('experimentConfigurationChanged'), 
                            self.__plotManager.onExpConfChanged)
            self.__plotManager.removePanels()
            self.__plotManager.setModel(None)
            self.__plotManager = None
            

def demo(model=None):
    """Experiment configuration"""
    #w = main_ChannelEditor()
    w = ExpDescriptionEditor()
    if model is None:
        from sardana.taurus.qt.qtgui.extra_macroexecutor import \
            TaurusMacroConfigurationDialog
        dialog = TaurusMacroConfigurationDialog(w)
        accept = dialog.exec_()
        if accept:
            model = str(dialog.doorComboBox.currentText())
    if model is not None:
        w.setModel(model)
    return w

def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="Exp. Description demo", app_version="1.0",
                          org_domain="Sardana", org_name="Tango community")

    args = app.get_command_line_args()
    if len(args) == 1:
        w = demo(model=args[0])
    else:
        w = demo()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
model.py: 
"""

from taurus.external.qt import Qt
from lxml import etree

from sardana.taurus.core.tango.sardana import macro


class MacrosListModel(Qt.QAbstractListModel):

    def __init__(self, parent=None):
        Qt.QAbstractListModel.__init__(self, parent)
        self.list = []

    def rowCount(self, parent=Qt.QModelIndex()):
        return len(self.list)

    def data(self, index, role):
        if index.isValid() and role == Qt.Qt.DisplayRole:
            macroNode = self.list[index.row()]
            return Qt.QVariant(self.list[index.row()].toSpockCommand())
        else:
            return Qt.QVariant()

    def index(self, row, column=0, parent=Qt.QModelIndex()):
        if self.rowCount():
            return self.createIndex(row, column, self.list[row])
        else:
            return Qt.QModelIndex()

    def insertRow(self, macroNode, row=0):
        self.beginInsertRows(Qt.QModelIndex(), row, row)
        self.list.insert(row, macroNode)
        self.endInsertRows()
        return self.index(row)

    def removeRow(self, row):
        self.beginRemoveRows(Qt.QModelIndex(), row, row)
        self.list.pop(row)
        self.endRemoveRows()
        if row == self.rowCount():
            row = row - 1
        return self.index(row)

    def isUpRowAllowed(self, index):
        return index.row() > 0

    def upRow(self, row):
        """This method move macro up and returns index with its new position"""
        macroNode = self.list[row]
        self.removeRow(row)
        return self.insertRow(macroNode, row - 1)

    def isDownRowAllowed(self, index):
        return index.row() < self.rowCount() - 1

    def downRow(self, row):
        """This method move macro down and returns index with its new position"""
        macroNode = self.list[row]
        self.removeRow(row)
        return self.insertRow(macroNode, row + 1)

    def toXmlString(self, pretty=False):
        listElement = etree.Element("list")
        for macroNode in self.list:
            listElement.append(macroNode.toXml(withId=False))
        xmlTree = etree.ElementTree(listElement)
        xmlString = etree.tostring(xmlTree, pretty_print=pretty)
        return xmlString

    def fromXmlString(self, xmlString):
        listElement = etree.fromstring(xmlString)
        for childElement in listElement.iterchildren("macro"):
            macroNode = macro.MacroNode()
            macroNode.fromXml(childElement)
            self.list.append(macroNode)
        self.reset()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of the macroserver base types for
macros"""

__all__ = ["Integer", "Float", "Boolean", "String", "User", "Filename",
           "File", "Macro", "MacroLibrary", "Env", "Motor", "MotorParam",
           "MotorGroup", "ExpChannel", "MeasurementGroup", "ComChannel",
           "IORegister", "Controller", "Instrument", "ControllerClass" ]

__docformat__ = 'restructuredtext'

from sardana import INTERFACES
from sardana.macroserver.msparameter import ParamType, AttrParamType, \
    ElementParamInterface

# Basic types
class Any(ParamType):
    type_class = lambda p : p

class Integer(ParamType):
    type_class = int

class Float(ParamType):
    type_class = float

class Boolean(ParamType):
    type_class = bool

    def getObj(self, str_repr):
        return str_repr.lower() == "true"
    
class String(ParamType):
    type_class = str

class User(ParamType):
    type_class = str

class Filename(ParamType):
    type_class = str

class File(ParamType):
    type_class = str
    
    def __init__(self, macro_server, name):
        ParamType.__init__(self, macro_server, name)
        self.filename = None
        #self.data is supposed to be an array.array object
        self.data = None
    
    def set(self, filename, data):
        self.filename = filename
        self.data = data

class JSON(ParamType):

    def getObj(self, str_repr):
        import json
        return json.loads(str_repr)
    
class Env(ParamType):
    type_class = str

class TangoDevice(ParamType):

    def getObj(self, str_repr):
        import PyTango
        return PyTango.DeviceProxy(str_repr)

class Device(ParamType):

    def getObj(self, str_repr):
        import taurus
        return taurus.Device(str_repr)
 
# Hardware types

class MotorParam(AttrParamType):
    """Class designed to represent a motor parameter name. Usual values
    are acceleration,deceleration,velocity,backlash,steps_per_unit,etc"""
    
    def __init__(self, macro_server, name):
        AttrParamType.__init__(self, macro_server, name)
        self.attr_item_list = ["Acceleration","Backlash","Base_rate","Step_per_unit",
                "Deceleration","Velocity","Offset"]
        self.non_attr_item_list = ["Controller"]
    
    def getItemList(self):
        return self.non_attr_item_list + self.attr_item_list
    
    def getAttrItemList(self):
        return self.attr_item_list
    
    def getNonAttrItemList(self):
        return self.non_attr_item_list


def __build_base_types():
    for sardana_type, info in INTERFACES.items():
        _, doc = info
        class _I(ElementParamInterface):
            __doc__ = doc
            __name__ = sardana_type
        globals()[sardana_type] = _I

__build_base_types()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines the base classes
for Sardana attributes"""

from __future__ import absolute_import

__all__ = ["SardanaAttribute", "SardanaSoftwareAttribute",
           "ScalarNumberAttribute", "SardanaAttributeConfiguration"]

__docformat__ = 'restructuredtext'

import weakref
import datetime

from .sardanaevent import EventGenerator, EventType
from .sardanadefs import ScalarNumberFilter
from .sardanavalue import SardanaValue


class SardanaAttribute(EventGenerator):
    """Class representing an atomic attribute like position of a motor or a
    counter value"""

    def __init__(self, obj, name=None, initial_value=None, **kwargs):
        super(SardanaAttribute, self).__init__(**kwargs)
        if obj is not None:
            obj = weakref.ref(obj)
        self._obj = obj
        self.name = name or self.__class__.__name__
        self._r_value = None
        self._last_event_value = None
        self._w_value = None
        self.filter = lambda a, b: True
        self.config = SardanaAttributeConfiguration()
        if initial_value is not None:
            self.set_value(initial_value)

    def has_value(self):
        """Determines if the attribute's read value has been read at least once
        in the lifetime of the attribute.
        
        :return: True if the attribute has a read value stored or False otherwise
        :rtype: bool"""
        return self._has_value()

    def _has_value(self):
        return not self._r_value is None

    def has_write_value(self):
        """Determines if the attribute's write value has been read at least once
        in the lifetime of the attribute.
        
        :return: True if the attribute has a write value stored or False otherwise
        :rtype: bool"""
        return self._has_write_value()

    def _has_write_value(self):
        return self._w_value is not None

    def get_obj(self):
        """Returns the object which *owns* this attribute
        
        :return: the object which *owns* this attribute
        :rtype: obj"""
        return self._get_obj()

    def _get_obj(self):
        obj = self._obj
        if obj is not None:
            obj = obj()
        return obj

    def in_error(self):
        """Determines if this attribute is in error state.
        
        :return: True if the attribute is in error state or False otherwise
        :rtype: bool"""
        return self._in_error()

    def _in_error(self):
        return self.has_value() and self._r_value.error

    def set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        """Sets the current read value and propagates the event (if
        propagate > 0).
        
        :param value: the new read value for this attribute. If a SardanaValue
                      is given, exc_info and timestamp are ignored (if given)
        :type value: obj or SardanaValue
        :param exc_info: exception information as returned by
                         :func:`sys.exc_info` [default: None, meaning no
                         exception]
        :type exc_info: tuple<3> or None
        :param timestamp: timestamp of attribute readout [default: None, meaning
                          create a 'now' timestamp]
        :type timestamp: float or None
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        return self._set_value(value, exc_info=exc_info, timestamp=timestamp,
                               propagate=propagate)

    def _set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        if isinstance(value, SardanaValue):
            rvalue = value
        else:
            rvalue = SardanaValue(value=value, exc_info=exc_info, timestamp=timestamp)
        self._r_value = rvalue
        self.fire_read_event(propagate=propagate)

    def get_value(self):
        """Returns the last read value for this attribute.
        
        :return: the last read value for this attribute
        :rtype: obj
        
        :raises: :exc:`Exception` if no read value has been set yet"""
        return self._get_value()

    def _get_value(self):
        return self.get_value_obj().value

    def get_value_obj(self):
        """Returns the last read value for this attribute.
        
        :return: the last read value for this attribute
        :rtype: :class:`~sardana.sardanavalue.SardanaValue`
        
        :raises: :exc:`Exception` if no read value has been set yet"""
        return self._get_value_obj()

    def _get_value_obj(self):
        if not self.has_value():
            raise Exception("{0}.{1} doesn't have a read value yet".format(
                            self.obj.name, self.name))
        return self._r_value

    def set_write_value(self, w_value, timestamp=None, propagate=1):
        """Sets the current write value.
        
        :param value: the new write value for this attribute. If a SardanaValue
                      is given, timestamp is ignored (if given)
        :type value: obj or SardanaValue
        :param timestamp: timestamp of attribute write [default: None, meaning
                          create a 'now' timestamp]
        :type timestamp: float or None
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        return self._set_write_value(w_value, timestamp=timestamp,
                                     propagate=propagate)

    def _set_write_value(self, w_value, timestamp=None, propagate=1):
        if isinstance(w_value, SardanaValue):
            wvalue = w_value
        else:
            wvalue = SardanaValue(value=w_value, timestamp=timestamp)
        self._w_value = wvalue
        self.fire_write_event(propagate=propagate)

    def get_write_value(self):
        """Returns the last write value for this attribute.
        
        :return: the last write value for this attribute or None if value has
                 not been written yet
        :rtype: obj"""
        return self._get_write_value()

    def _get_write_value(self):
        w_value = self.get_write_value_obj()
        if w_value is not None:
            return w_value.value

    def get_write_value_obj(self):
        """Returns the last write value object for this attribute.
        
        :return: the last write value for this attribute or None if value has
                 not been written yet
        :rtype: :class:`~sardana.sardanavalue.SardanaValue`"""
        return self._get_write_value_obj()

    def _get_write_value_obj(self):
        if self.has_write_value():
            return self._w_value

    def get_exc_info(self):
        """Returns the exception information (like :func:`sys.exc_info`) about
        last attribute readout or None if last read did not generate an
        exception.
        
        :return: exception information or None
        :rtype: tuple<3> or None"""
        return self._get_exc_info()

    def _get_exc_info(self):
        exc_info = None
        if self.has_value():
            exc_info = self._r_value.exc_info
        return exc_info

    def accepts(self, propagate):
        if propagate < 1:
            return False
        if self._last_event_value is None:
            return True
        return propagate > 1 or self.filter(self.get_value(), self._last_event_value.value)

    def get_timestamp(self):
        """Returns the timestamp of the last readout or None if the attribute 
        has never been read before
        
        :return: timestamp of the last readout or None
        :rtype: float or None"""
        return self._get_timestamp()

    def _get_timestamp(self):
        if self.has_value():
            return self._r_value.timestamp

    def get_write_timestamp(self):
        """Returns the timestamp of the last write or None if the attribute 
        has never been written before
        
        :return: timestamp of the last write or None
        :rtype: float or None"""
        return self._get_write_timestamp()

    def _get_write_timestamp(self):
        if self.has_write_value():
            return self._w_value.timestamp

    def fire_write_event(self, propagate=1):
        """Fires an event to the listeners of the object which owns this
        attribute.
        
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        if propagate < 1:
            return
        evt_type = EventType("w_" + self.name, priority=propagate)
        self.fire_event(evt_type, self)

    def fire_read_event(self, propagate=1):
        """Fires an event to the listeners of the object which owns this
        attribute.
        
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        if self.accepts(propagate):
            obj = self.obj
            if obj is not None:
                self._last_event_value = self._r_value
                evt_type = EventType(self.name, priority=propagate)
                self.fire_event(evt_type, self)

    obj = property(get_obj, "container object for this attribute")
    value_obj = property(get_value_obj)
    write_value_obj = property(get_write_value_obj)
    value = property(get_value, set_value, "current read value for this attribute")
    w_value = property(get_write_value, set_write_value,
                       "current write value for this attribute")
    timestamp = property(get_timestamp, doc="the read timestamp")
    w_timestamp = property(get_write_timestamp, doc="the write timestamp")
    error = property(in_error)
    exc_info = property(get_exc_info)

    def __repr__(self):
        v = None
        if self.in_error():
            v = "<Error>"
        elif self.has_value():
            v = self.value
        return "{0}(value={1})".format(self.name, v)

    def __str__(self):
        if self.has_value():
            value = "{0} at {1}".format(self.value, datetime.datetime.fromtimestamp(self.timestamp))
        else:
            value = "-----"
        if self.has_write_value():
            w_value = "{0} at {1}".format(self.w_value, datetime.datetime.fromtimestamp(self.w_timestamp))
        else:
            w_value = "-----"

        ret = """{0.__class__.__name__}(
       name = {0.name}
    manager = {0.obj}
       read = {1}
      write = {2})
""".format(self, value, w_value)
        return ret


class SardanaSoftwareAttribute(SardanaAttribute):
    """Class representing a software attribute. The difference between this and
    :class:`SardanaAttribute` is that, because it is a pure software attribute,
    there is no difference ever between the read and write values."""

    get_value = SardanaAttribute.get_value

    def set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        """Sets the current read value and propagates the event (if
        propagate > 0).
        
        :param value: the new read value for this attribute
        :type value: obj
        :param exc_info: exception information as returned by
                         :func:`sys.exc_info` [default: None, meaning no
                         exception]
        :type exc_info: tuple<3> or None
        :param timestamp: timestamp of attribute readout [default: None, meaning
                          create a 'now' timestamp]
        :type timestamp: float or None
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int"""
        SardanaAttribute.set_value(self, value, exc_info=exc_info,
                                   timestamp=timestamp)
        self.set_write_value(value, timestamp=self.timestamp)

    value = property(get_value, set_value, "current value for this attribute")


class ScalarNumberAttribute(SardanaAttribute):
    """A :class:`SardanaAttribute` specialized for numbers"""

    def __init__(self, *args, **kwargs):
        SardanaAttribute.__init__(self, *args, **kwargs)
        self.filter = ScalarNumberFilter()


class SardanaAttributeConfiguration(object):
    """Storage class for :class:`SardanaAttribute` information (like ranges)"""
    NoRange = float('-inf'), float('inf')

    def __init__(self):
        self.range = self.NoRange
        self.alarm = self.NoRange
        self.warning = self.NoRange




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains sardana base wizard classes."""

__all__ = ["SardanaBaseWizard"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt


class SardanaBasePage(Qt.QWizardPage):

    def __init__(self, parent=None):
        Qt.QWizardPage.__init__(self, parent)
        self._item_funcs = {}

    def __setitem__(self, name, value):
        self._item_funcs[name] = value

    def __getitem__(self, name):
        return self._item_funcs[name]

    def getPanelWidget(self):
        return self._panel


class SardanaBaseWizard(Qt.QWizard):

    def __init__(self, parent=None):
        Qt.QWizard.__init__(self, parent)
        self._item_funcs = {}
        self._pages = {}

    def __setitem__(self, name, value):
        self._item_funcs[name] = value

    def __getitem__(self, name):
        for id in self.getPages():
            p = self.page(id)
            if isinstance(p, SardanaBasePage):
                try:
                    return p[name]()
                except Exception, e:
                    pass
        return self._item_funcs[name]()
        return None

    def addPage(self, page):
        id = Qt.QWizard.addPage(self, page)
        self._pages[id] = page

    def setPage(self, id, page):
        Qt.QWizard.setPage(self, id, page)
        self._pages[id] = page

    def getPages(self):
        return self._pages




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from taurus.external.qt import Qt
from taurus import Database
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core.taurusdatabase import TaurusAttrInfo
from taurus.qt.qtgui.input import TaurusAttrListComboBox
from taurus.qt.qtgui.tree import TaurusDbTreeWidget
from taurus.qt.qtgui.resource import getThemeIcon
from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.macroparameterseditor import MacroParametersEditor
from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.parameditors import LineEditParam, ParamBase, ComboBoxParam, CheckBoxParam, DirPathParam, MSAttrListComboBoxParam
from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.model import ParamEditorModel
from sardana.taurus.qt.qtgui.extra_macroexecutor.common import MSAttrListComboBox


class SenvEditor(Qt.QWidget, MacroParametersEditor):

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        MacroParametersEditor.__init__(self)
        self.valueWidget = None

    def initComponents(self):
        self.setLayout(Qt.QFormLayout())

        self.layout().addRow(Qt.QLabel("Setting environment variable:", self))

        self.nameComboBox = ComboBoxParam(self)
        self.nameComboBox.addItems(["ActiveMntGrp", "ExtraColumns", "JsonRecorder", "ScanFile", "ScanDir"])
        self.nameComboBox.setEditable(True)
        self.connect(self.nameComboBox, Qt.SIGNAL("currentIndexChanged(int)"), self.onNameComboBoxChanged)
        self.layout().addRow("name:", self.nameComboBox)

        nameIndex = self.model().index(0, 1, self.rootIndex())
        self.nameComboBox.setIndex(nameIndex)

    def setRootIndex(self, rootIndex):
        self._rootIndex = rootIndex
        self.initComponents()

    def rootIndex(self):
        return self._rootIndex

    def model(self):
        return self._model

    def setModel(self, model):
        self._model = model
        if isinstance(model, ParamEditorModel):
            self.setRootIndex(Qt.QModelIndex())

    def onNameComboBoxChanged(self, index):
        text = str(self.nameComboBox.currentText())
        if self.valueWidget is not None:
            label = self.layout().labelForField(self.valueWidget)
            if label is not None:
                self.layout().removeWidget(label)
                label.setParent(None)
                label = None

            self.layout().removeWidget(self.valueWidget)
            self.valueWidget.resetValue()
            self.valueWidget.setParent(None)
            self.valueWidget = None

        self.valueWidget, label = getSenvValueEditor(text, self)

        paramRepeatIndex = self.model().index(1, 0, self.rootIndex())
        repeatIndex = paramRepeatIndex.child(0, 0)
        valueIndex = repeatIndex.child(0, 1)
        self.valueWidget.setIndex(valueIndex)

        if label:
            self.layout().addRow(label, self.valueWidget)
        else:
            self.layout().addRow(self.valueWidget)

def getSenvValueEditor(envName, parent):
    """Factory method, requires: string, and QWidget as a parent for returned editor.
    Factory returns a tuple of widget and a label for it.
    
    :return: (Qt.QWidget, str) """
    label = "value:"
    if envName == "ActiveMntGrp":
        editor = MSAttrListComboBoxParam(parent)
        editor.setUseParentModel(True)
        editor.setModel("/MeasurementGroupList")
    elif envName == "ExtraColumns":
        editor = ExtraColumnsEditor(parent)
        label = None
    elif envName == "JsonRecorder":
        editor = CheckBoxParam(parent)
    elif envName == "ScanDir":
        editor = DirPathParam(parent)
    elif envName == "ScanFile":
        editor = LineEditParam(parent)
    else:
        editor = LineEditParam(parent)
    return editor, label

class ExtraColumnsEditor(ParamBase, Qt.QWidget):

    def __init__(self, parent=None, paramModel=None):
        ParamBase.__init__(self, paramModel)
        Qt.QWidget.__init__(self, parent)
        self.setLayout(Qt.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        addNewColumnButton = Qt.QPushButton(getThemeIcon("list-add") , "Add new column...", self)
        removeSelectedColumnsButton = Qt.QPushButton(getThemeIcon("list-remove") , "Remove selected...", self)
        buttonsLayout = Qt.QHBoxLayout()
        buttonsLayout.addWidget(addNewColumnButton)
        buttonsLayout.addWidget(removeSelectedColumnsButton)
        self.layout().addLayout(buttonsLayout)

        self.extraColumnsTable = ExtraColumnsTable(self)
        self.extraColumnsModel = ExtraColumnsModel()
        self.extraColumnsTable.setModel(self.extraColumnsModel)
        self.extraColumnsTable.setItemDelegate(ExtraColumnsDelegate(self.extraColumnsTable))

        self.layout().addWidget(self.extraColumnsTable)

        self.connect(addNewColumnButton, Qt.SIGNAL("clicked()"), self.onAddNewColumn)
        self.connect(removeSelectedColumnsButton, Qt.SIGNAL("clicked()"), self.onRemoveSelectedColumns)
        self.connect(self.extraColumnsModel, Qt.SIGNAL("dataChanged (const QModelIndex&,const QModelIndex&)"), self.onExtraColumnsChanged)
        self.connect(self.extraColumnsModel, Qt.SIGNAL("modelReset()"), self.onExtraColumnsChanged)

    def getValue(self):
        return repr(self.extraColumnsTable.model().columns())

    def setValue(self, value):
        try:
            columns = eval(value)
        except:
            columns = []
        self.extraColumnsTable.setColumns(columns)

    def onAddNewColumn(self):
        self.extraColumnsTable.insertRows()
        self.emit(Qt.SIGNAL("modelChanged()"))

    def onRemoveSelectedColumns(self):
        self.extraColumnsTable.removeRows()
        self.emit(Qt.SIGNAL("modelChanged()"))

    def onExtraColumnsChanged(self):
        self.emit(Qt.SIGNAL("modelChanged()"))


class ExtraColumnsTable(Qt.QTableView):

    def __init__(self, parent):
        Qt.QTableView.__init__(self, parent)
        self.setSelectionBehavior(Qt.QAbstractItemView.SelectRows)
        self.setSelectionMode(Qt.QAbstractItemView.ExtendedSelection)

    def setColumns(self, columns):
        if columns == None: columns = []
        self.model().setColumns(columns)
        self.resizeColumnsToContents()

    def insertRows(self):
        self.model().insertRows(self.model().rowCount())

    def removeRows(self):
        rows = [index.row() for index in self.selectedIndexes()]
        rows = list(set(rows))
        rows.sort(reverse=True)
        for row in rows:
            self.model().removeRows(row)


class ExtraColumnsDelegate(Qt.QItemDelegate):

    def __init__(self, parent=None):
        Qt.QItemDelegate.__init__(self, parent)
        db = Database()
        self.host = db.getNormalName()

    def createEditor(self, parent, option, index):
        if index.column() == 1:
            self.combo_attr_tree_widget = TaurusDbTreeWidget(perspective=TaurusElementType.Device)
            self.combo_attr_tree_widget.setModel(self.host)
            treeView = self.combo_attr_tree_widget.treeView()
            qmodel = self.combo_attr_tree_widget.getQModel()
            editor = Qt.QComboBox(parent)
            editor.setModel(qmodel)
            editor.setMaxVisibleItems(20)
            editor.setView(treeView)
        elif index.column() == 2:
            editor = MSAttrListComboBox(parent)
            editor.setUseParentModel(True)
            editor.setModel("/InstrumentList")
        else:
            editor = Qt.QItemDelegate.createEditor(self, parent, option, index)
        return editor

    def setEditorData(self, editor, index):
        if index.column() == 2:
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            editor.setCurrentText(text)
        else:
            Qt.QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        column = index.column()
        if column == 1:
            selectedItems = self.combo_attr_tree_widget.selectedItems()
            if not len(selectedItems) == 1: return
            taurusTreeAttributeItem = selectedItems[0]
            itemData = taurusTreeAttributeItem.itemData()
            if isinstance(itemData, TaurusAttrInfo):
                model.setData(index, Qt.QVariant(itemData.fullName()))
        elif column == 2:
            model.setData(index, Qt.QVariant(editor.currentText()))
        else:
            Qt.QItemDelegate.setModelData(self, editor, model, index)

    def sizeHint(self, option, index):
        if index.column() == 0:
            fm = option.fontMetrics
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            document = Qt.QTextDocument()
            document.setDefaultFont(option.font)
            document.setHtml(text)
            size = Qt.QSize(document.idealWidth() + 5, fm.height())
        elif index.column() == 1:
            editor = self.createEditor(self.parent(), option, index)
            if editor is None:
                size = Qt.QItemDelegate.sizeHint(self, option, index)
            else:
                size = editor.sizeHint()
                editor.hide()
                editor.setParent(None)
#                editor.destroy()
        else:
            size = Qt.QItemDelegate.sizeHint(self, option, index)
        return size

class ExtraColumnsModel(Qt.QAbstractTableModel):


    def __init__(self, columns=None):
        if columns is None: columns = []
        Qt.QAbstractItemModel.__init__(self)
        self.__columns = columns

    def setColumns(self, columns):
        self.__columns = columns
        self.reset()

    def columns(self):
        return self.__columns

    def rowCount(self, index=Qt.QModelIndex()):
        return len(self.__columns)

    def columnCount(self, index=Qt.QModelIndex()):
        return 3

    def data(self, index, role=Qt.Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return Qt.QVariant()
        row = index.row()
        column = index.column()
        #Display Role
        if role == Qt.Qt.DisplayRole:
            if column == 0: return Qt.QVariant(Qt.QString(self.__columns[row]['label']))
            elif column == 1: return Qt.QVariant(Qt.QString(self.__columns[row]['model']))
            elif column == 2: return Qt.QVariant(Qt.QString(self.__columns[row]['instrument']))
        return Qt.QVariant()

    def headerData(self, section, orientation, role=Qt.Qt.DisplayRole):
        if role == Qt.Qt.TextAlignmentRole:
            if orientation == Qt.Qt.Horizontal:
                return Qt.QVariant(int(Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter))
            return Qt.QVariant(int(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter))
        if role != Qt.Qt.DisplayRole:
            return Qt.QVariant()
        #So this is DisplayRole...
        if orientation == Qt.Qt.Horizontal:
            if section == 0: return Qt.QVariant("Label")
            elif section == 1: return Qt.QVariant("Attribute")
            elif section == 2: return Qt.QVariant("Instrument")
            return Qt.QVariant()
        else:
            return Qt.QVariant(Qt.QString.number(section + 1))

    def flags(self, index):
        flags = Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsSelectable
        if index.isValid():
            column = index.column()
            if column in (0, 1, 2):
                flags |= Qt.Qt.ItemIsEditable
        return flags

    def setData(self, index, value=None, role=Qt.Qt.EditRole):
        if index.isValid() and (0 <= index.row() < self.rowCount()):
            row = index.row()
            column = index.column()
            value = Qt.from_qvariant(value, str)
            if column == 0: self.__columns[row]['label'] = value
            elif column == 1: self.__columns[row]['model'] = value
            elif column == 2: self.__columns[row]['instrument'] = value
            self.emit(Qt.SIGNAL("dataChanged(QModelIndex,QModelIndex)"), index, index)
            return True
        return False

    def insertRows(self, row, rows=1, parentindex=None):
        if parentindex is None: parentindex = Qt.QModelIndex()
        first = row
        last = row + rows - 1
        self.beginInsertRows(parentindex, first, last)
        for row in range(first, last + 1):
            self.insertRow(row)
        self.endInsertRows()
        return True

    def insertRow(self, row, parentIndex=None):
        self.__columns.insert(row, {'label':'', 'model':'', 'instrument':''})

    def removeRows(self, row, rows=1, parentindex=None):
        if parentindex is None: parentindex = Qt.QModelIndex()
        first = row
        last = row + rows - 1
        self.beginRemoveRows(parentindex, first, last)
        for row in range(first, last + 1):
            self.removeRow(row)
        self.endRemoveRows()
        return True

    def removeRow(self, row, parentIndex=None):
        self.__columns.pop(row)

CUSTOM_EDITOR = SenvEditor

if __name__ == "__main__":
    import sys
    import taurus
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)
    args = app.get_command_line_args()
    editor = SenvEditor()
    macroServer = taurus.Device(args[0])
    macroInfoObj = macroServer.getMacroInfoObj("senv")
    macroNode = MacroNode()
    editor.setMacroNode(macroNode)
    editor.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
dooroutput.py: 
"""

import taurus.core
from taurus.external.qt import Qt


class DoorOutput(Qt.QPlainTextEdit):
    """Widget used for displaying changes of door's attributes: Output, Info, Warning and Error."""

    def __init__(self, parent=None):
        Qt.QTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))
        self.stopAction = Qt.QAction("Stop scrolling", self)
        self.stopAction.setCheckable(True)
        self.stopAction.setChecked(False)
        self._isStopped = False

    def onDoorOutputChanged(self, output):
        """call on output attribute changed"""
        txt = "<font color=\"Black\">"
        if output is None:
            return
        for line in output:
            txt += "OUTPUT  " + line.replace(' ', '&nbsp;') + "\n"
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorInfoChanged(self, info):
        """call on info attribute changed"""
        txt = "<font color=\"Black\">"
        if info is None:
            return

        for line in info:
            txt += "INFO  " + line.replace(' ', '&nbsp;') + "\n"
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorWarningChanged(self, warning):
        """call on warning attribute changed"""
        txt = "<font color=\"Black\">"
        if warning is None:
            return
        for line in warning:
            txt += "WARNING  " + line.replace(' ', '&nbsp;') + "\n"
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorErrorChanged(self, error):
        """call on error attribute changed"""
        txt = "<font color=\"Red\">"
        if error is None:
            return
        for line in error:
            txt += "ERROR  " + line.replace(' ', '&nbsp;') + "\n"
        txt += "</font>"
        self.appendHtmlText(txt)

    def appendHtmlText(self, text):
        self.appendHtml(text)
        if not self._isStopped:
            self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        menu.addAction(self.stopAction)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        Qt.QObject.connect(clearAction, Qt.SIGNAL("triggered()"), self.clear)
        Qt.QObject.connect(self.stopAction, Qt.SIGNAL("toggled(bool)"), self.stopScrolling)
        menu.exec_(event.globalPos())

    def stopScrolling(self, stop):
        self._isStopped = stop

class DoorDebug(Qt.QPlainTextEdit):
    """Widget used for displaying changes of door's Debug attribute."""

    def __init__(self, parent=None):
        Qt.QTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))
        self.stopAction = Qt.QAction("Stop scrolling", self)
        self.stopAction.setCheckable(True)
        self.stopAction.setChecked(False)
        self._isStopped = False

    def onDoorDebugChanged(self, debug):
        """call on debug attribute changed"""
        if debug is None:
            return
        for line in debug:
            self.appendPlainText("DEBUG  " + line)

        if not self._isStopped:
            self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        menu.addAction(self.stopAction)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        Qt.QObject.connect(clearAction, Qt.SIGNAL("triggered()"), self.clear)
        Qt.QObject.connect(self.stopAction, Qt.SIGNAL("toggled(bool)"), self.stopScrolling)
        menu.exec_(event.globalPos())

    def stopScrolling(self, stop):
        self._isStopped = stop

class DoorResult(Qt.QPlainTextEdit):
    """Widget used for displaying changes of door's Result attribute."""

    def __init__(self, parent=None):
        Qt.QTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))

    def onDoorResultChanged(self, result):
        """call on result attribute changed"""
        if result is None:
            return
        for line in result:
            self.appendPlainText("RESULT  " + line)
        self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        Qt.QObject.connect(clearAction, Qt.SIGNAL("triggered()"), self.clear)
        menu.exec_(event.globalPos())



#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Door attributes listeners
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

class DoorAttrListener(Qt.QObject):

    def __init__(self, attrName):
        Qt.QObject.__init__(self)
        self.attrName = attrName
        self.attrObj = None

    def setDoorName(self, doorName):
        if not self.attrObj is None:
            self.attrObj.removeListener(self)
        self.attrObj = taurus.Attribute(doorName, self.attrName)
        self.attrObj.addListener(self)

    def eventReceived(self, src, type, value):
        if (type == taurus.core.taurusbasetypes.TaurusEventType.Error or
            type == taurus.core.taurusbasetypes.TaurusEventType.Config):
            return
        self.emit(Qt.SIGNAL('door%sChanged' % self.attrName), value.value)

if __name__ == "__main__":
    import sys
    import taurus
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)
    args = app.get_command_line_args()

    doorOutput = DoorOutput()
    if len(args) == 1:
        door = taurus.Device(args[0])
        Qt.QObject.connect(door, Qt.SIGNAL("outputUpdated"), doorOutput.onDoorOutputChanged)
        Qt.QObject.connect(door, Qt.SIGNAL("infoUpdated"), doorOutput.onDoorInfoChanged)
        Qt.QObject.connect(door, Qt.SIGNAL("warningUpdated"), doorOutput.onDoorWarningChanged)
        Qt.QObject.connect(door, Qt.SIGNAL("errorUpdated"), doorOutput.onDoorErrorChanged)
    doorOutput.show()
    sys.exit(app.exec_())




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for OneDExpChannel"""

__all__ = ["Pool1DExpChannel"]

__docformat__ = 'restructuredtext'

from sardana import ElementType
from sardana.sardanaevent import EventType

from sardana.pool.poolbasechannel import PoolBaseChannel


class Pool1DExpChannel(PoolBaseChannel):

    def __init__(self, **kwargs):
        self._data_source = None
        kwargs['elem_type'] = ElementType.OneDExpChannel
        PoolBaseChannel.__init__(self, **kwargs)

    # -------------------------------------------------------------------------
    # data source
    # -------------------------------------------------------------------------

    def get_data_source(self, cache=True, propagate=1):
        if not cache or self._data_source is None:
            data_source = self.read_data_source()
            self._set_data_source(data_source, propagate=propagate)
        return self._data_source

    def _set_data_source(self, data_source, propagate=1):
        self._data_source = data_source
        if not propagate:
            return
        self.fire_event(EventType("data_source", priority=propagate),
                        data_source)

    def read_data_source(self):
        data_source = self.controller.get_axis_par(self.axis, "data_source")
        return data_source

    data_source = property(get_data_source,
                           doc="source identifier for the 1D data")



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
delegate.py: 
"""

from taurus.external.qt import Qt

from sardana.taurus.qt.qtgui.extra_macroexecutor import globals
from sardana.taurus.core.tango.sardana import macro

from .parameditors import MSAttrListComboBoxParam, SpinBoxParam, \
    DoubleSpinBoxParam, LineEditParam, FileDialogParam, ComboBoxParam, ComboBoxBoolean

class ParamEditorDelegate(Qt.QStyledItemDelegate):

    def __init__(self, parent=None):
        Qt.QStyledItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        if index.column() == 1:
            node = index.model().nodeFromIndex(index)
            if isinstance(node, macro.SingleParamNode):
                paramType = node.type()
                if paramType in globals.EDITOR_COMBOBOX_PARAMS:
                    comboBox = MSAttrListComboBoxParam(parent, node)
                    comboBox.setElementType(paramType)

                    ##################
                    # The setUseParentModel mechanism is not working
                    # we do it manually here as a hack
                    #comboBox.setUseParentModel(True)
                    #comboBox.setModel('/elements')
                    w = parent
                    while w is not None:
                        if hasattr(w, 'getModelName'):
                            model = w.getModelName() + '/elements'
                            break
                        w = w.parent()
                    comboBox.setModel(model)
                    ###################

                    return comboBox
                elif paramType in globals.EDITOR_SPINBOX_PARAMS:
                    return SpinBoxParam(parent, node)
                elif paramType in globals.EDITOR_DOUBLESPINBOX_PARAMS:
                    return DoubleSpinBoxParam(parent, node)
                elif paramType in globals.EDITOR_LINEEDIT_PARAMS:
                    return LineEditParam(parent, node)
                elif paramType in globals.EDITOR_FILEDIALOG_PARAMS:
                    return FileDialogParam(parent, node)
                elif paramType in globals.EDITOR_BOOLEAN_PARAMS:
                    return ComboBoxBoolean(parent, node)
        return Qt.QStyledItemDelegate.createEditor(self, parent, option, index)

    def setEditorData(self, editor, index):
        if index.column() == 1:
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            if text == "None" or text == "":
                Qt.QStyledItemDelegate.setEditorData(self, editor, index)
            else:
                editor.setValue(text)
#                node = index.model().nodeFromIndex(index)
#                paramType = node.type()
#                if paramType in globals.EDITOR_COMBOBOX_PARAMS :
#                    editor.setValue(text)
#                elif paramType in globals.EDITOR_SPINBOX_PARAMS:
#                    editor.setValue(int(text))
#                elif paramType in globals.EDITOR_DOUBLESPINBOX_PARAMS:
#                    editor.setValue(float(text))
#                elif paramType in globals.EDITOR_LINEEDIT_PARAMS:
#                    editor.setText(text)
#                elif paramType in globals.EDITOR_FILEDIALOG_PARAMS:
#                    editor.filePath.setText(text)
        else:
            Qt.QStyledItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        if index.column() == 1:
            model.setData(index, Qt.QVariant(editor.getValue()))
        else:
            Qt.QStyledItemDelegate.setModelData(self, editor, model, index)

    def sizeHint(self, option, index):
        if index.column() == 0:
            fm = option.fontMetrics
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            document = Qt.QTextDocument()
            document.setDefaultFont(option.font)
            document.setHtml(text)
            size = Qt.QSize(document.idealWidth() + 5, fm.height())
        elif index.column() == 1:
            editor = self.createEditor(self.parent(), option, index)
            if editor is None:
                size = Qt.QStyledItemDelegate.sizeHint(self, option, index)
            else:
                size = editor.sizeHint()
                editor.hide()
                editor.setParent(None)
#                editor.destroy()
        else:
            size = Qt.QStyledItemDelegate.sizeHint(self, option, index)
        return size



#!/usr/bin/env python
from sardana.pool.poolextension import ControllerStateTranslator

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__docformat__ = 'restructuredtext'


def prepare_pool(util):

    import PyTango
    from sardana.pool.poolextension import ControllerValueTranslator, \
        register_controller_value_translator, \
        ControllerStateTranslator, register_controller_state_translator, \
        CannotTranslateException
    from sardana.tango.core.util import from_deviceattribute

    class TangoControllerValueTranslator(ControllerValueTranslator):

        def translate(self, value):
            if not isinstance(value, PyTango.DeviceAttribute):
                return super(TangoControllerValueTranslator, self).translate(value)
            ret = from_deviceattribute(value)
            return ret

    register_controller_value_translator(TangoControllerValueTranslator)

    class TangoControllerStateTranslator(ControllerStateTranslator):

        def translate(self, value):
            if not isinstance(value, PyTango.DeviceAttribute):
                return super(TangoControllerValueTranslator, self).translate(value)
            if value.type != PyTango.DevState:
                raise CannotTranslateException("Expected DevState got %s" % value.type)
            ret = from_deviceattribute(value)
            return ret

    register_controller_state_translator(TangoControllerStateTranslator)

    from .Controller import ControllerClass, Controller
    from .Motor import MotorClass, Motor
    from .PseudoMotor import PseudoMotorClass, PseudoMotor
    from .MotorGroup import MotorGroupClass, MotorGroup
    from .CTExpChannel import CTExpChannelClass, CTExpChannel
    from .ZeroDExpChannel import ZeroDExpChannelClass, ZeroDExpChannel
    from .OneDExpChannel import OneDExpChannelClass, OneDExpChannel
    from .TwoDExpChannel import TwoDExpChannelClass, TwoDExpChannel
    from .PseudoCounter import PseudoCounterClass, PseudoCounter
    from .MeasurementGroup import MeasurementGroupClass, MeasurementGroup
    from .IORegister import IORegisterClass, IORegister
    from .Pool import PoolClass, Pool

    util.add_class(PoolClass, Pool)
    util.add_class(ControllerClass, Controller)
    util.add_class(MotorClass, Motor)
    util.add_class(IORegisterClass, IORegister)
    util.add_class(CTExpChannelClass, CTExpChannel)
    util.add_class(ZeroDExpChannelClass, ZeroDExpChannel)
    util.add_class(OneDExpChannelClass, OneDExpChannel)
    util.add_class(TwoDExpChannelClass, TwoDExpChannel)
    util.add_class(PseudoMotorClass, PseudoMotor)
    util.add_class(PseudoCounterClass, PseudoCounter)
    util.add_class(MotorGroupClass, MotorGroup)
    util.add_class(MeasurementGroupClass, MeasurementGroup)

def main_pool(args=None, start_time=None, mode=None):
    import sardana.tango.core.util
    return sardana.tango.core.util.run(prepare_pool, args=args,
                                       start_time=start_time, mode=mode)

run = main_pool



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Spock submodule. It contains an input handler"""

__all__ = ['SpockInputHandler', 'InputHandler']

__docformat__ = 'restructuredtext'

import sys
from multiprocessing import Process, Pipe

from taurus.core import TaurusManager
from taurus.core.util.singleton import Singleton
from taurus.external.qt import Qt
from taurus.qt.qtgui.dialog import TaurusMessageBox, TaurusInputDialog

from sardana.taurus.core.tango.sardana.macroserver import BaseInputHandler

from sardana.spock import genutils


class SpockInputHandler(BaseInputHandler):

    def __init__(self):
        # don't call super __init__ on purpose
        self._input = genutils.spock_input

    def input(self, input_data=None):
        if input_data is None:
            input_data = {}
        prompt = input_data.get('prompt')
        ret = dict(input=None, cancel=False)
        try:
            if prompt is None:
                ret['input'] = self._input()
            else:
                ret['input'] = self._input(prompt)
        except:
            ret['cancel'] = True
        return ret

    def input_timeout(self, input_data):
        print "SpockInputHandler input timeout"


class MessageHandler(Qt.QObject):

    def __init__(self, conn, parent=None):
        Qt.QObject.__init__(self, parent)
        self._conn = conn
        self._dialog = None
        self.connect(self, Qt.SIGNAL("messageArrived"), self.on_message)
        
    def handle_message(self, input_data):
        self.emit(Qt.SIGNAL("messageArrived"), input_data)
    
    def on_message(self, input_data):
        msg_type = input_data['type']
        if msg_type == 'input':
            if 'macro_name' in input_data and 'title' not in input_data:
                input_data['title'] = input_data['macro_name']
            self._dialog = dialog = TaurusInputDialog(input_data=input_data)
            dialog.activateWindow()
            dialog.exec_()
            ok = dialog.result()
            value = dialog.value()
            ret = dict(input=None, cancel=False)
            if ok:
                ret['input'] = value
            else:
                ret['cancel'] = True
            self._conn.send(ret)
        elif msg_type == 'timeout':
            dialog = self._dialog
            if dialog:
                dialog.close()


class InputHandler(Singleton, BaseInputHandler):
    
    def __init__(self):
        # don't call super __init__ on purpose
        pass
            
    def init(self, *args, **kwargs):
        self._conn, child_conn = Pipe()
        self._proc = proc = Process(target=self.safe_run,
            name="SpockInputHandler", args=(child_conn,))
        proc.daemon = True
        proc.start()
    
    def input(self, input_data=None):
        # parent process
        data_type = input_data.get('data_type', 'String')
        if isinstance(data_type, (str, unicode)):
            ms = genutils.get_macro_server()
            interfaces = ms.getInterfaces()
            if data_type in interfaces:
                input_data['data_type'] = [ elem.name for elem in interfaces[data_type].values() ]
        self._conn.send(input_data)
        ret = self._conn.recv()
        return ret

    def input_timeout(self, input_data):
        # parent process
        self._conn.send(input_data)
        
    def safe_run(self, conn):
        # child process
        try:
            return self.run(conn)
        except Exception, e:
            msgbox = TaurusMessageBox(*sys.exc_info())
            conn.send((e, False))
            msgbox.exec_()
    
    def run(self, conn):
        # child process
        self._conn = conn
        app = Qt.QApplication.instance()
        if app is None:
            app = Qt.QApplication([])
        app.setQuitOnLastWindowClosed(False)
        self._msg_handler = MessageHandler(conn)
        TaurusManager().addJob(self.run_forever, None)
        app.exec_()
        conn.close()
        print "Quit input handler"
                
    def run_forever(self):
        # child process
        message, conn = True, self._conn
        while message:
            message = conn.recv()
            if not message:
                continue
            self._msg_handler.handle_message(message)
        app = Qt.QApplication.instance()
        if app:
            app.quit()




import PyTango
import taurus
import csv
import sys

def checkPoolElements(pool):
    pool_dev = taurus.Device(pool)

    # GET CONTROLLER CLASSES
    pool_ctrl_classes = {}
    for info in pool_dev['ControllerClassList'].value:
        ctrl_class, library_path, ctrl_type = info.split()
        pool_ctrl_classes[ctrl_class] = (library_path, ctrl_type)

    # GET CONTROLLERS INFO
    pool_controllers_by_type = {'Motor': [],
                                'PseudoMotor': [],
                                'IORegister': [],
                                'CounterTimer': [],
                                'PseudoCounter': [],
                                'ZeroDExpChannel': []}
    pool_controllers = {}
    for info in pool_dev['ControllerList'].value:
        info_splitted = info.split()
        ctrl_name = info_splitted[0]
        ctrl_library = info_splitted[-1].replace('(','').replace(')','')
        ctrl_class = info_splitted[1].rsplit('.')[1].split('/')[0]
        ctrl_type = ''
        try:
            ctrl_type = pool_ctrl_classes[ctrl_class][1]
            pool_controllers_by_type[ctrl_type].append(ctrl_name)
        except:
            pass
        ctrl_properties = {}
        ctrl_elements = []
        ctrl_pool_elements = []
        for prop in pool_dev.get_property_list(ctrl_name+'/*'):
            prop_name = prop.split('/')[1]
            prop_value = pool_dev.get_property(prop)[prop]
            ctrl_properties[prop_name] = ' '.join(prop_value)
        pool_controllers[ctrl_name] = {}
        pool_controllers[ctrl_name]['type'] = ctrl_type
        pool_controllers[ctrl_name]['name'] = ctrl_name
        pool_controllers[ctrl_name]['file'] = ctrl_library
        pool_controllers[ctrl_name]['class'] = ctrl_class
        pool_controllers[ctrl_name]['properties'] = ctrl_properties
        pool_controllers[ctrl_name]['elements'] = ctrl_elements
        pool_controllers[ctrl_name]['ctrl_pool_elements'] = ctrl_pool_elements

    for ctrl_type, controllers in pool_controllers_by_type.iteritems():
        if len(controllers) == 0:
            continue
        
        print '\n'
        print '----------------------------------------------------------------'
        print len(controllers), ctrl_type, 'CONTROLLERS FOR POOL '+pool
        print '----------------------------------------------------------------'

        for ctrl in controllers:
            details = pool_controllers[ctrl]
            ###print 'Controller',ctrl,':'
            ###for k, v in details.iteritems():
            ###    if k == 'ctrl_pool_elements':
            ###        print '\tElements count:\t'+str(len(v))
            ###    print '\t'+k+':\t'+str(v)
            ###
            ###pool_controllers[ctrl_name]['type'] = ctrl_type
            ###pool_controllers[ctrl_name]['name'] = ctrl_name
            ###pool_controllers[ctrl_name]['file'] = ctrl_library
            ###pool_controllers[ctrl_name]['class'] = ctrl_class
            ###pool_controllers[ctrl_name]['properties'] = ctrl_properties
            ###pool_controllers[ctrl_name]['elements'] = ctrl_elements
            ###pool_controllers[ctrl_name]['ctrl_pool_elements'] = ctrl_pool_elements

            pool_elements_summary = '('+str(len(details['ctrl_pool_elements']))+')'
            print '{type}\t{name}\t{file}\t{class}\t{properties}\t{elements}'.format(**details), pool_elements_summary
            

    pool_elements = {}
    pool_elements['ExpChannels'] = pool_dev['ExpChannelList'].value or []
    pool_elements['Motors'] = pool_dev['MotorList'].value or []
    pool_elements['IORegs'] = pool_dev['IORegisterList'].value or []
    pool_instruments = pool_dev['InstrumentList'].value or []

    print '\n'
    print '----------------------------------------------------------------'
    print len(pool_instruments),'INSTRUMENTS FOR POOL '+pool
    print '----------------------------------------------------------------'
    print pool_instruments

    # CHECK ELEMENTS WITHOUT INSTRUMENT
    for element_type in pool_elements.keys():
        elements = pool_elements[element_type]
        elements_with_no_instrument = []
        for info in elements:
            info_splitted = info.split()
            alias = info_splitted[0]
            dev_name = info_splitted[1]
            ctrl_name, ctrl_axis = info_splitted[2].replace('(','').replace(')','').split('/')
            specific_element_type = info_splitted[3]

            pool_controllers[ctrl_name]['ctrl_pool_elements'].append(alias)
            if specific_element_type in ['PseudoMotor', 'PseudoCounter']:
                if len(pool_controllers[ctrl_name]['elements']) == 0:
                    pool_controllers[ctrl_name]['elements'] = ' '.join(info_splitted[4:])

            element_dev = taurus.Device(alias)
            if element_dev['Instrument'].value == '':
                elements_with_no_instrument.append(alias)
        if len(elements_with_no_instrument) > 0:
            print '\n***',element_type,'with no Instrument:',elements_with_no_instrument,'***'

    print '\n'
    print '----------------------------------------------------------------'
    print 'ELEMENTS FOR POOL '+pool
    print '----------------------------------------------------------------'
    for element_type in pool_elements.keys():
        elements = pool_elements[element_type]
        print element_type+':',len(elements)

    print '\n'
    print '----------------------------------------------------------------'
    print 'ELEMENTS MEMORIZED ATTRIBUTES AND ATTRIBUTE CONFIGURATIONS FOR POOL '+pool
    print '----------------------------------------------------------------'
    db = taurus.Database()
    for element_type in pool_elements.keys():
        elements = pool_elements[element_type]
        for info in elements:
            info_splitted = info.split()
            alias = info_splitted[0]
            dev_name = info_splitted[1]
            ctrl_name, ctrl_axis = info_splitted[2].replace('(','').replace(')','').split('/')
            specific_element_type = info_splitted[3]
            element_dev = taurus.Device(alias)
            for attr, attr_dict in db.get_device_attribute_property(element_dev.getNormalName(), element_dev.get_attribute_list()).iteritems():
                if len(attr_dict) > 0:
                    print specific_element_type, alias, attr, attr_dict
                else:
                    if attr.lower() in ['position', 'value']:
                        print '***',specific_element_type, alias, attr,  'NO MEMORIZED ATTRIBUTES OR ATTRIBUTE CONFIGURATIONS ***'




if __name__ == '__main__':
    if len(sys.argv) != 2 or sys.argv[1] == '?':
        print '----------------------------------------'
        print 'Invalid number of arguments.'
        print ''
        print 'Example of usage:'
        print '    python get_pool_config pool'
        print ''
        print '    where pool is the name of the pool'
        print '----------------------------------------'


    pool = sys.argv[1]
    checkPoolElements(pool)
